"""
Abstraction of the ice in a cryo-EM image.
"""

from abc import abstractmethod
from typing_extensions import override

import jax.numpy as jnp
import jax.random as jr
from equinox import Module
from jaxtyping import Array, Complex, PRNGKeyArray

from ..image.operators import FourierOperatorLike
from ._instrument_config import InstrumentConfig
from ._scattering_theory import compute_phase_shifts_from_integrated_potential


class AbstractIce(Module, strict=True):
    """Base class for an ice model."""

    @abstractmethod
    def sample_fourier_phase_shifts_from_ice(
        self, key: PRNGKeyArray, instrument_config: InstrumentConfig
    ) -> Complex[
        Array, "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}"
    ]:
        """Sample a stochastic realization of the phase shifts due to the ice
        at the exit plane."""
        raise NotImplementedError

    def compute_fourier_phase_shifts_with_ice(
        self,
        key: PRNGKeyArray,
        fourier_phase_at_exit_plane: Complex[
            Array,
            "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}",
        ],
        instrument_config: InstrumentConfig,
    ) -> Complex[
        Array, "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}"
    ]:
        """Compute the combined phase of the ice and the specimen."""
        # Sample the realization of the phase due to the ice.
        fourier_ice_phase_at_exit_plane = self.sample_fourier_phase_shifts_from_ice(
            key, instrument_config
        )

        return fourier_phase_at_exit_plane + fourier_ice_phase_at_exit_plane


class GaussianIce(AbstractIce, strict=True):
    r"""Ice modeled as gaussian noise.

    **Attributes:**

    - `variance_function` :
        A function that computes the variance
        of the ice, modeled as colored gaussian noise.
        The dimensions of this function are the square
        of the dimensions of an integrated potential.
    """

    variance_function: FourierOperatorLike

    def __init__(self, variance_function: FourierOperatorLike):
        self.variance_function = variance_function

    @override
    def sample_fourier_phase_shifts_from_ice(
        self, key: PRNGKeyArray, instrument_config: InstrumentConfig
    ) -> Complex[
        Array, "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}"
    ]:
        """Sample a realization of the ice phase shifts as colored gaussian noise."""
        n_pixels = instrument_config.padded_n_pixels
        frequency_grid_in_angstroms = instrument_config.padded_frequency_grid_in_angstroms
        # Compute standard deviation, scaling up by the variance by the number
        # of pixels to make the realization independent pixel-independent in real-space.
        std = jnp.sqrt(n_pixels * self.variance_function(frequency_grid_in_angstroms))
        ice_integrated_potential_at_exit_plane = std * jr.normal(
            key,
            shape=frequency_grid_in_angstroms.shape[0:-1],
            dtype=complex,
        ).at[0, 0].set(0.0)
        ice_phase_shifts_at_exit_plane = compute_phase_shifts_from_integrated_potential(
            ice_integrated_potential_at_exit_plane,
            instrument_config.wavelength_in_angstroms,
        )

        return ice_phase_shifts_at_exit_plane



from abc import abstractmethod
from typing import Optional

import equinox as eqx
from jaxtyping import Array, Complex, PRNGKeyArray

from .._instrument_config import InstrumentConfig


class AbstractScatteringTheory(eqx.Module, strict=True):
    """Base class for a scattering theory."""

    @abstractmethod
    def compute_fourier_contrast_at_detector_plane(
        self,
        instrument_config: InstrumentConfig,
        rng_key: Optional[PRNGKeyArray] = None,
    ) -> Complex[
        Array, "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}"
    ]:
        raise NotImplementedError

    @abstractmethod
    def compute_fourier_squared_wavefunction_at_detector_plane(
        self,
        instrument_config: InstrumentConfig,
        rng_key: Optional[PRNGKeyArray] = None,
    ) -> Complex[
        Array, "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}"
    ]:
        raise NotImplementedError



"""
Representations of conformational variables.
"""

from typing import Any

from equinox import AbstractVar, Module


class AbstractConformationalVariable(Module, strict=True):
    """A conformational variable wrapped in an `equinox.Module`."""

    value: AbstractVar[Any]


AbstractConformationalVariable.__init__.__doc__ = """**Arguments:**

- `value`: The value of the integer conformation.
"""



"""
Image formation models simulated from gaussian noise distributions.
"""

from typing import Optional
from typing_extensions import override

import jax.numpy as jnp
import jax.random as jr
from equinox import field
from jaxtyping import Array, Complex, Float, PRNGKeyArray

from ..._errors import error_if_not_positive
from ...image import normalize_image, rfftn
from ...image.operators import Constant, FourierOperatorLike
from ...simulator import AbstractImagingPipeline
from ._base_distribution import AbstractDistribution


class IndependentGaussianFourierModes(AbstractDistribution, strict=True):
    r"""A gaussian noise model, where each fourier mode is independent.

    This computes the likelihood in Fourier space,
    so that the variance to be an arbitrary noise power spectrum.
    """

    imaging_pipeline: AbstractImagingPipeline
    variance_function: FourierOperatorLike
    signal_scale_factor: Float[Array, ""]

    is_signal_normalized: bool = field(static=True)

    def __init__(
        self,
        imaging_pipeline: AbstractImagingPipeline,
        variance_function: Optional[FourierOperatorLike] = None,
        signal_scale_factor: Optional[float | Float[Array, ""]] = None,
        is_signal_normalized: bool = False,
    ):
        """**Arguments:**

        - `imaging_pipeline`: The image formation model.
        - `variance_function`: The variance of each fourier mode. By default,
                               `cryojax.image.operators.Constant(1.0)`.
        - `signal_scale_factor`: A scale factor for the underlying signal simulated from `imaging_pipeline`.
        - `is_signal_normalized`:
            Whether or not the signal is normalized before applying the `signal_scale_factor`.
            If an `AbstractMask` is given to `imaging_pipeline.mask`, the signal is normalized
            within the region where the mask is equal to `1`.
        """  # noqa: E501
        self.imaging_pipeline = imaging_pipeline
        self.variance_function = variance_function or Constant(1.0)
        if signal_scale_factor is None:
            signal_scale_factor = jnp.sqrt(
                jnp.asarray(imaging_pipeline.instrument_config.n_pixels, dtype=float)
            )
        self.signal_scale_factor = error_if_not_positive(jnp.asarray(signal_scale_factor))
        self.is_signal_normalized = is_signal_normalized

    @override
    def compute_signal(
        self, *, get_real: bool = True
    ) -> (
        Float[
            Array,
            "{self.imaging_pipeline.instrument_config.y_dim} "
            "{self.imaging_pipeline.instrument_config.x_dim}",
        ]
        | Complex[
            Array,
            "{self.imaging_pipeline.instrument_config.y_dim}"
            " {self.imaging_pipeline.instrument_config.x_dim//2+1}",
        ]
    ):
        """Render the image formation model."""
        if self.is_signal_normalized:
            simulated_image = self.imaging_pipeline.render(get_real=True)
            where = (
                None
                if self.imaging_pipeline.mask is None
                else self.imaging_pipeline.mask.array == 1.0
            )
            normalized_simulated_image = normalize_image(simulated_image, where=where)
            return (
                self.signal_scale_factor * normalized_simulated_image
                if get_real
                else self.signal_scale_factor * rfftn(normalized_simulated_image)
            )
        else:
            real_or_fourier_simulated_image = self.imaging_pipeline.render(
                get_real=get_real
            )
            return self.signal_scale_factor * real_or_fourier_simulated_image

    def compute_noise(
        self, rng_key: PRNGKeyArray, *, get_real: bool = True
    ) -> (
        Float[
            Array,
            "{self.imaging_pipeline.instrument_config.y_dim} "
            "{self.imaging_pipeline.instrument_config.x_dim}",
        ]
        | Complex[
            Array,
            "{self.imaging_pipeline.instrument_config.y_dim} "
            "{self.imaging_pipeline.instrument_config.x_dim//2+1}",
        ]
    ):
        pipeline = self.imaging_pipeline
        n_pixels = pipeline.instrument_config.padded_n_pixels
        freqs = pipeline.instrument_config.padded_frequency_grid_in_angstroms
        # Compute the zero mean variance and scale up to be independent of the number of
        # pixels
        std = jnp.sqrt(n_pixels * self.variance_function(freqs))
        noise = pipeline.postprocess(
            std
            * jr.normal(rng_key, shape=freqs.shape[0:-1])
            .at[0, 0]
            .set(0.0)
            .astype(complex),
            get_real=get_real,
        )

        return noise

    @override
    def sample(
        self, rng_key: PRNGKeyArray, *, get_real: bool = True
    ) -> (
        Float[
            Array,
            "{self.imaging_pipeline.instrument_config.y_dim} "
            "{self.imaging_pipeline.instrument_config.x_dim}",
        ]
        | Complex[
            Array,
            "{self.imaging_pipeline.instrument_config.y_dim} "
            "{self.imaging_pipeline.instrument_config.x_dim//2+1}",
        ]
    ):
        """Sample from the gaussian noise model."""
        return self.compute_signal(get_real=get_real) + self.compute_noise(
            rng_key, get_real=get_real
        )

    @override
    def log_likelihood(
        self,
        observed: Complex[
            Array,
            "{self.imaging_pipeline.instrument_config.y_dim} "
            "{self.imaging_pipeline.instrument_config.x_dim//2+1}",
        ],
    ) -> Float[Array, ""]:
        """Evaluate the log-likelihood of the gaussian noise model.

        **Arguments:**

        - `observed` : The observed data in fourier space.
        """
        pipeline = self.imaging_pipeline
        n_pixels = pipeline.instrument_config.n_pixels
        freqs = pipeline.instrument_config.frequency_grid_in_angstroms
        # Compute the variance and scale up to be independent of the number of pixels
        variance = n_pixels * self.variance_function(freqs)
        # Create simulated data
        simulated = self.compute_signal(get_real=False)
        # Compute residuals
        residuals = simulated - observed
        # Compute standard normal random variables
        squared_standard_normal_per_mode = jnp.abs(residuals) ** 2 / (2 * variance)
        # Compute the log-likelihood for each fourier mode.
        log_likelihood_per_mode = (
            squared_standard_normal_per_mode - jnp.log(2 * jnp.pi * variance) / 2
        )
        # Compute log-likelihood, throwing away the zero mode. Need to take care
        # to compute the loss function in fourier space for a real-valued function.
        log_likelihood = (
            -1.0
            * (
                jnp.sum(log_likelihood_per_mode[1:, 0])
                + 2 * jnp.sum(log_likelihood_per_mode[:, 1:])
            )
            / n_pixels
        )

        return log_likelihood



"""
Voxel-based representations of the scattering potential.
"""

from abc import abstractmethod
from functools import cached_property
from typing import cast, ClassVar, Optional
from typing_extensions import override, Self

import equinox as eqx
import jax.numpy as jnp
import numpy as np
from equinox import AbstractClassVar, AbstractVar, field
from jaxtyping import Array, Complex, Float

from ..._errors import error_if_not_positive
from ...coordinates import make_coordinate_grid, make_frequency_slice
from ...image import (
    compute_spline_coefficients,
    crop_to_shape,
    fftn,
    pad_to_shape,
)
from ...image.operators import AbstractFilter
from .._pose import AbstractPose
from .base_potential import AbstractPotentialRepresentation


class AbstractVoxelPotential(AbstractPotentialRepresentation, strict=True):
    """Abstract interface for a voxel-based scattering potential representation."""

    voxel_size: AbstractVar[Float[Array, ""]]
    is_real: AbstractClassVar[bool]

    @property
    @abstractmethod
    def shape(self) -> tuple[int, ...]:
        """The shape of the voxel array."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_real_voxel_grid(
        cls,
        real_voxel_grid: Float[Array, "dim dim dim"] | Float[np.ndarray, "dim dim dim"],
        voxel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
    ) -> Self:
        """Load an `AbstractVoxelPotential` from real-valued 3D electron
        scattering potential.
        """
        raise NotImplementedError


# Not public API
class AbstractFourierVoxelGridPotential(AbstractVoxelPotential, strict=True):
    """Abstract interface of a 3D scattering potential voxel grid
    in fourier-space.
    """

    frequency_slice_in_pixels: AbstractVar[Float[Array, "1 dim dim 3"]]

    @abstractmethod
    def __init__(
        self,
        fourier_voxel_grid: Complex[Array, "dim dim dim"],
        frequency_slice_in_pixels: Float[Array, "1 dim dim 3"],
        voxel_size: Float[Array, ""] | float,
    ):
        raise NotImplementedError

    @cached_property
    def frequency_slice_in_angstroms(self) -> Float[Array, "1 dim dim 3"]:
        """The `frequency_slice_in_pixels` in angstroms."""
        return _safe_multiply_grid_by_constant(
            self.frequency_slice_in_pixels, 1 / self.voxel_size
        )

    def rotate_to_pose(self, pose: AbstractPose) -> Self:
        """Return a new potential with a rotated `frequency_slice_in_pixels`."""
        return eqx.tree_at(
            lambda d: d.frequency_slice_in_pixels,
            self,
            pose.rotate_coordinates(self.frequency_slice_in_pixels, inverse=True),
        )

    @classmethod
    def from_real_voxel_grid(
        cls,
        real_voxel_grid: Float[Array, "dim dim dim"] | Float[np.ndarray, "dim dim dim"],
        voxel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
        *,
        pad_scale: float = 1.0,
        pad_mode: str = "constant",
        filter: Optional[AbstractFilter] = None,
    ) -> Self:
        """Load an `AbstractFourierVoxelGridPotential` from real-valued 3D electron
        scattering potential voxel grid.

        **Arguments:**

        - `real_voxel_grid`: A scattering potential voxel grid in real space.
        - `voxel_size`: The voxel size of `real_voxel_grid`.
        - `pad_scale`: Scale factor at which to pad `real_voxel_grid` before fourier
                     transform. Must be a value greater than `1.0`.
        - `pad_mode`: Padding method. See `jax.numpy.pad` for documentation.
        - `filter`: A filter to apply to the result of the fourier transform of
                  `real_voxel_grid`, i.e. `fftn(real_voxel_grid)`. Note that the zero
                  frequency component is assumed to be in the corner.
        """
        # Cast to jax array
        real_voxel_grid, voxel_size = (
            jnp.asarray(real_voxel_grid),
            jnp.asarray(voxel_size),
        )
        # Pad template
        if pad_scale < 1.0:
            raise ValueError("`pad_scale` must be greater than 1.0")
        # ... always pad to even size to avoid interpolation issues in
        # fourier slice extraction.
        padded_shape = cast(
            tuple[int, int, int],
            tuple([int(s * pad_scale) for s in real_voxel_grid.shape]),
        )
        padded_real_voxel_grid = pad_to_shape(
            real_voxel_grid, padded_shape, mode=pad_mode
        )
        # Load potential and coordinates. For now, do not store the
        # fourier potential only on the half space. Fourier slice extraction
        # does not currently work if rfftn is used.
        fourier_voxel_grid_with_zero_in_corner = (
            fftn(padded_real_voxel_grid)
            if filter is None
            else filter(fftn(padded_real_voxel_grid))
        )
        # ... store the potential grid with the zero frequency component in the center
        fourier_voxel_grid = jnp.fft.fftshift(fourier_voxel_grid_with_zero_in_corner)
        # ... create in-plane frequency slice on the half space
        frequency_slice = make_frequency_slice(
            cast(tuple[int, int], padded_real_voxel_grid.shape[:-1]), half_space=False
        )

        return cls(fourier_voxel_grid, frequency_slice, voxel_size)


class FourierVoxelGridPotential(AbstractFourierVoxelGridPotential):
    """A 3D scattering potential voxel grid in fourier-space."""

    fourier_voxel_grid: Complex[Array, "dim dim dim"]
    frequency_slice_in_pixels: Float[Array, "1 dim dim 3"]
    voxel_size: Float[Array, ""] = field(converter=error_if_not_positive)

    is_real: ClassVar[bool] = False

    @override
    def __init__(
        self,
        fourier_voxel_grid: Complex[Array, "dim dim dim"],
        frequency_slice_in_pixels: Float[Array, "1 dim dim 3"],
        voxel_size: Float[Array, ""] | float,
    ):
        """**Arguments:**

        - `fourier_voxel_grid`: The cubic voxel grid in fourier space.
        - `frequency_slice_in_pixels`: Frequency slice coordinate system.
        - `voxel_size`: The voxel size.
        """
        self.fourier_voxel_grid = jnp.asarray(fourier_voxel_grid)
        self.frequency_slice_in_pixels = frequency_slice_in_pixels
        self.voxel_size = jnp.asarray(voxel_size)

    @property
    def shape(self) -> tuple[int, int, int]:
        """The shape of the `fourier_voxel_grid`."""
        return cast(tuple[int, int, int], self.fourier_voxel_grid.shape)


class FourierVoxelGridPotentialInterpolator(AbstractFourierVoxelGridPotential):
    """A 3D scattering potential voxel grid in fourier-space, represented
    by spline coefficients.
    """

    coefficients: Float[Array, "coeff_dim coeff_dim coeff_dim"]
    frequency_slice_in_pixels: Float[Array, "1 dim dim 3"]
    voxel_size: Float[Array, ""] = field(converter=error_if_not_positive)

    is_real: ClassVar[bool] = False

    def __init__(
        self,
        fourier_voxel_grid: Float[Array, "dim dim dim"],
        frequency_slice_in_pixels: Float[Array, "1 dim dim 3"],
        voxel_size: Float[Array, ""] | float,
    ):
        """
        !!! note
            The argument `fourier_voxel_grid` is used to set
            `FourierVoxelGridPotentialInterpolator.coefficients` in the `__init__`,
            but it is not stored in the class. For example,

            ```python
            voxels = FourierVoxelGridPotentialInterpolator(
                fourier_voxel_grid, frequency_slice, voxel_size
            )
            assert hasattr(voxels, "fourier_voxel_grid")  # This will return an error
            assert hasattr(voxels, "coefficients")  # Instead, store spline coefficients
            ```

        **Arguments:**

        - `fourier_voxel_grid`: The cubic voxel grid in fourier space.
        - `frequency_slice_in_pixels`: Frequency slice coordinate system,
                                               wrapped in a `FrequencySlice` object.
        - `voxel_size`: The voxel size.
        """
        self.coefficients = compute_spline_coefficients(jnp.asarray(fourier_voxel_grid))
        self.frequency_slice_in_pixels = frequency_slice_in_pixels
        self.voxel_size = jnp.asarray(voxel_size)

    @property
    def shape(self) -> tuple[int, int, int]:
        """The shape of the original `fourier_voxel_grid` from which
        `coefficients` were computed.
        """
        return cast(tuple[int, int, int], tuple([s - 2 for s in self.coefficients.shape]))


class RealVoxelGridPotential(AbstractVoxelPotential, strict=True):
    """Abstraction of a 3D scattering potential voxel grid in real-space."""

    real_voxel_grid: Float[Array, "dim dim dim"]
    coordinate_grid_in_pixels: Float[Array, "dim dim dim 3"]
    voxel_size: Float[Array, ""] = field(converter=error_if_not_positive)

    is_real: ClassVar[bool] = True

    def __init__(
        self,
        real_voxel_grid: Float[Array, "dim dim dim"],
        coordinate_grid_in_pixels: Float[Array, "dim dim dim 3"],
        voxel_size: Float[Array, ""] | float,
    ):
        """**Arguments:**

        - `real_voxel_grid`: The voxel grid in fourier space.
        - `coordinate_grid_in_pixels`: A coordinate grid.
        - `voxel_size`: The voxel size.
        """
        self.real_voxel_grid = jnp.asarray(real_voxel_grid)
        self.coordinate_grid_in_pixels = coordinate_grid_in_pixels
        self.voxel_size = jnp.asarray(voxel_size)

    @property
    def shape(self) -> tuple[int, int, int]:
        """The shape of the `real_voxel_grid`."""
        return cast(tuple[int, int, int], self.real_voxel_grid.shape)

    @cached_property
    def coordinate_grid_in_angstroms(self) -> Float[Array, "dim dim dim 3"]:
        """The `coordinate_grid_in_pixels` in angstroms."""
        return _safe_multiply_grid_by_constant(
            self.coordinate_grid_in_pixels, self.voxel_size
        )

    def rotate_to_pose(self, pose: AbstractPose) -> Self:
        """Return a new potential with a rotated
        `coordinate_grid_in_pixels`.
        """
        return eqx.tree_at(
            lambda d: d.coordinate_grid_in_pixels,
            self,
            pose.rotate_coordinates(self.coordinate_grid_in_pixels, inverse=False),
        )

    @classmethod
    def from_real_voxel_grid(
        cls,
        real_voxel_grid: Float[Array, "dim dim dim"] | Float[np.ndarray, "dim dim dim"],
        voxel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
        *,
        coordinate_grid_in_pixels: Optional[Float[Array, "dim dim dim 3"]] = None,
        crop_scale: Optional[float] = None,
    ) -> Self:
        """Load a `RealVoxelGridPotential` from a real-valued 3D electron
        scattering potential voxel grid.

        **Arguments:**

        - `real_voxel_grid`: An electron scattering potential voxel grid in real space.
        - `voxel_size`: The voxel size of `real_voxel_grid`.
        - `crop_scale`: Scale factor at which to crop `real_voxel_grid`.
                        Must be a value greater than `1`.
        """
        # Cast to jax array
        real_voxel_grid, voxel_size = (
            jnp.asarray(real_voxel_grid),
            jnp.asarray(voxel_size),
        )
        # Make coordinates if not given
        if coordinate_grid_in_pixels is None:
            # Option for cropping template
            if crop_scale is not None:
                if crop_scale < 1.0:
                    raise ValueError("`crop_scale` must be greater than 1.0")
                cropped_shape = cast(
                    tuple[int, int, int],
                    tuple([int(s / crop_scale) for s in real_voxel_grid.shape[-3:]]),
                )
                real_voxel_grid = crop_to_shape(real_voxel_grid, cropped_shape)
            coordinate_grid_in_pixels = make_coordinate_grid(real_voxel_grid.shape[-3:])

        return cls(real_voxel_grid, coordinate_grid_in_pixels, voxel_size)


class RealVoxelCloudPotential(AbstractVoxelPotential, strict=True):
    """Abstraction of a 3D electron scattering potential voxel point cloud.

    !!! info
        This object is similar to the `RealVoxelGridPotential`. Instead
        of storing the whole voxel grid, a `RealVoxelCloudPotential` need
        only store points of non-zero scattering potential. Therefore,
        a `RealVoxelCloudPotential` stores a point cloud of scattering potential
        voxel values. Instantiating with the `from_real_voxel_grid` constructor
        will automatically mask points of zero scattering potential.
    """

    voxel_weights: Float[Array, " size"]
    coordinate_list_in_pixels: Float[Array, "size 3"]
    voxel_size: Float[Array, ""] = field(converter=error_if_not_positive)

    is_real: ClassVar[bool] = True

    def __init__(
        self,
        voxel_weights: Float[Array, " size"],
        coordinate_list_in_pixels: Float[Array, "size 3"],
        voxel_size: Float[Array, ""] | float,
    ):
        """**Arguments:**

        - `voxel_weights`: A point-cloud of voxel scattering potential values.
        - `coordinate_list_in_pixels`: Coordinate list for the `voxel_weights`.
        - `voxel_size`: The voxel size.
        """
        self.voxel_weights = jnp.asarray(voxel_weights)
        self.coordinate_list_in_pixels = coordinate_list_in_pixels
        self.voxel_size = jnp.asarray(voxel_size)

    @property
    def shape(self) -> tuple[int]:
        """The shape of `voxel_weights`."""
        return cast(tuple[int], self.voxel_weights.shape)

    @cached_property
    def coordinate_list_in_angstroms(self) -> Float[Array, "size 3"]:
        """The `coordinate_list_in_pixels` in angstroms."""
        return _safe_multiply_list_by_constant(
            self.coordinate_list_in_pixels, self.voxel_size
        )

    def rotate_to_pose(self, pose: AbstractPose) -> Self:
        """Return a new potential with a rotated
        `coordinate_list_in_pixels`.
        """
        return eqx.tree_at(
            lambda d: d.coordinate_list_in_pixels,
            self,
            pose.rotate_coordinates(self.coordinate_list_in_pixels, inverse=False),
        )

    @classmethod
    def from_real_voxel_grid(
        cls,
        real_voxel_grid: Float[Array, "dim dim dim"] | Float[np.ndarray, "dim dim dim"],
        voxel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
        *,
        coordinate_grid_in_pixels: Optional[Float[Array, "dim dim dim 3"]] = None,
        rtol: float = 1e-05,
        atol: float = 1e-08,
        size: Optional[int] = None,
        fill_value: Optional[float] = None,
    ) -> Self:
        """Load an `RealVoxelCloudPotential` from a real-valued 3D electron
        scattering potential voxel grid.

        **Arguments:**

        - `real_voxel_grid`: An electron scattering potential voxel grid in real space.
        - `voxel_size`: The voxel size of `real_voxel_grid`.
        - `rtol`: Argument passed to `jnp.isclose`, used for masking
                  voxels of zero scattering potential.
        - `atol`: Argument passed to `jnp.isclose`, used for masking
                  voxels of zero scattering potential.
        - `size`: Argument passed to `jnp.where`, used for fixing the size
                  of the masked scattering potential. This argument is required
                  for using this function with a JAX transformation.
        - `fill_value`: Argument passed to `jnp.where`, used if `size` is specified and
                        the mask has fewer than the indicated number of elements.
        """
        # Cast to jax array
        real_voxel_grid, voxel_size = (
            jnp.asarray(real_voxel_grid),
            jnp.asarray(voxel_size),
        )
        # Make coordinates if not given
        if coordinate_grid_in_pixels is None:
            coordinate_grid_in_pixels = make_coordinate_grid(real_voxel_grid.shape)
        # ... mask zeros to store smaller arrays. This
        # option is not jittable.
        nonzero = jnp.where(
            ~jnp.isclose(real_voxel_grid, 0.0, rtol=rtol, atol=atol),
            size=size,
            fill_value=fill_value,
        )
        flat_potential = real_voxel_grid[nonzero]
        coordinate_list = coordinate_grid_in_pixels[nonzero]

        return cls(flat_potential, coordinate_list, voxel_size)


def _safe_multiply_grid_by_constant(
    grid: Float[Array, "z_dim y_dim x_dim 3"], constant: Float[Array, ""]
) -> Float[Array, "z_dim y_dim x_dim 3"]:
    """Multiplies a coordinate grid by a constant in a
    safe way for gradient computation.
    """
    return jnp.where(grid != 0.0, jnp.asarray(constant) * grid, 0.0)


def _safe_multiply_list_by_constant(
    coordinate_list: Float[Array, "size 3"], constant: Float[Array, ""]
) -> Float[Array, "size 3"]:
    """Multiplies a coordinate grid by a constant in a
    safe way for gradient computation.
    """
    return jnp.where(coordinate_list != 0.0, jnp.asarray(constant) * coordinate_list, 0.0)



"""
Functions for creating coordinate systems.
"""

from typing import Optional

import jax.numpy as jnp
import numpy as np
from jaxtyping import Array, Float


def make_coordinate_grid(
    shape: tuple[int, ...],
    grid_spacing: float | Float[np.ndarray, ""] | Float[Array, ""] = 1.0,
) -> Float[Array, "*shape ndim"]:
    """
    Create a real-space cartesian coordinate system on a grid.

    **Arguments:**

    - `shape`: Shape of the voxel grid, with `ndim = len(shape)`.
    - `grid_spacing`: The grid spacing, in units of length.

    **Returns:**

    A cartesian coordinate system in real space.
    """
    coordinate_grid = _make_coordinates_or_frequencies(
        shape, grid_spacing=grid_spacing, real_space=True
    )
    return coordinate_grid


def make_frequency_grid(
    shape: tuple[int, ...],
    grid_spacing: float | Float[np.ndarray, ""] | Float[Array, ""] = 1.0,
    half_space: bool = True,
) -> Float[Array, "*shape ndim"]:
    """Create a fourier-space cartesian coordinate system on a grid.
    The zero-frequency component is in the corner.

    Arguments
    ---------
    - `shape`: Shape of the voxel grid, with `ndim = len(shape)`.
    - `grid_spacing`: The grid spacing, in units of length.
    - `half_space`:
        Return a frequency grid on the half space.
        `shape[-1]` is the axis on which the negative
        frequencies are omitted.

    **Returns:**

    A cartesian coordinate system in frequency space.
    """
    frequency_grid = _make_coordinates_or_frequencies(
        shape,
        grid_spacing=grid_spacing,
        real_space=False,
        half_space=half_space,
    )
    return frequency_grid


def make_frequency_slice(
    shape: tuple[int, int],
    grid_spacing: float | Float[np.ndarray, ""] | Float[Array, ""] = 1.0,
    half_space: bool = True,
) -> Float[Array, "1 {shape[0]} {shape[1]} 3"]:
    """Create a fourier-space cartesian coordinate system on a grid. The
    zero-frequency component is in the *center*. This is different
    than in [`make_frequency_grid`][].

    **Returns:**

    The central, $q_z = 0$ slice of a 3D frequency grid `$(q_x, q_y, q_z)$`.
    """
    frequency_slice = make_frequency_grid(shape, grid_spacing, half_space=half_space)
    if half_space:
        frequency_slice = jnp.fft.fftshift(frequency_slice, axes=(0,))
    else:
        frequency_slice = jnp.fft.fftshift(frequency_slice, axes=(0, 1))
    frequency_slice = jnp.expand_dims(
        jnp.pad(
            frequency_slice,
            ((0, 0), (0, 0), (0, 1)),
            mode="constant",
            constant_values=0.0,
        ),
        axis=0,
    )
    return frequency_slice


def make_1d_coordinate_grid(
    size: int,
    grid_spacing: float | Float[np.ndarray, ""] | Float[Array, ""] = 1.0,
) -> Float[Array, "*shape ndim"]:
    """
    Create a 1D real-space cartesian coordinate array.

    **Arguments:**

    - `size`: Size of the coordinate array.
    - `grid_spacing`: The grid spacing, in units of length.

    **Returns:**

    A 1D cartesian coordinate array in real space.
    """
    coordinate_array = _make_coordinates_or_frequencies_1d(
        size, grid_spacing=grid_spacing, real_space=True
    )
    return coordinate_array


def make_1d_frequency_grid(
    size: int,
    grid_spacing: float | Float[np.ndarray, ""] | Float[Array, ""] = 1.0,
    half_space: bool = True,
) -> Float[Array, "*shape ndim"]:
    """Create a 1D fourier-space cartesian coordinate array.
    If `half_space = False`, the zero-frequency component is in the beginning.

    Arguments
    ---------
    - `size`: Size of the coordinate array.
    - `grid_spacing`: The grid spacing, in units of length.
    - `half_space`:
        Return a frequency grid on the half space.
        `shape[-1]` is the axis on which the negative
        frequencies are omitted.

    **Returns:**

    A 1D cartesian coordinate array in frequency space.
    """
    frequency_array = _make_coordinates_or_frequencies_1d(
        size,
        grid_spacing=grid_spacing,
        real_space=False,
        rfftfreq=half_space,
    )
    return frequency_array


def _make_coordinates_or_frequencies(
    shape: tuple[int, ...],
    grid_spacing: float | Float[np.ndarray, ""] | Float[Array, ""] = 1.0,
    real_space: bool = False,
    half_space: bool = True,
) -> Float[Array, "*shape ndim"]:
    ndim = len(shape)
    coords1D = []
    for idx in range(ndim):
        if real_space:
            c1D = _make_coordinates_or_frequencies_1d(
                shape[idx], grid_spacing, real_space
            )
        else:
            if not half_space:
                rfftfreq = False
            else:
                rfftfreq = False if idx < ndim - 1 else True
            c1D = _make_coordinates_or_frequencies_1d(
                shape[idx], grid_spacing, real_space, rfftfreq
            )
        coords1D.append(c1D)
    if ndim == 2:
        y, x = coords1D
        xv, yv = jnp.meshgrid(x, y, indexing="xy")
        coords = jnp.stack([xv, yv], axis=-1)
    elif ndim == 3:
        z, y, x = coords1D
        xv, yv, zv = jnp.meshgrid(x, y, z, indexing="xy")
        xv, yv, zv = [
            jnp.transpose(rv, axes=[2, 0, 1]) for rv in [xv, yv, zv]
        ]  # Change axis ordering to [z, y, x]
        coords = jnp.stack([xv, yv, zv], axis=-1)
    else:
        raise ValueError(
            "Only 2D and 3D coordinate grids are supported. "
            f"Tried to create a grid of shape {shape}."
        )

    return coords


def _make_coordinates_or_frequencies_1d(
    size: int,
    grid_spacing: float | Float[np.ndarray, ""] | Float[Array, ""],
    real_space: bool = False,
    rfftfreq: Optional[bool] = None,
) -> Float[Array, " size"]:
    """One-dimensional coordinates in real or fourier space"""
    if real_space:
        make_1d = lambda size, dx: jnp.fft.fftshift(jnp.fft.fftfreq(size, 1 / dx)) * size
    else:
        if rfftfreq is None:
            raise ValueError("Argument rfftfreq cannot be None if real_space=False.")
        else:
            fn = jnp.fft.rfftfreq if rfftfreq else jnp.fft.fftfreq
            make_1d = lambda size, dx: fn(size, grid_spacing)

    return make_1d(size, grid_spacing)



from .base_transfer_theory import (
    AbstractTransferFunction as AbstractTransferFunction,
    AbstractTransferTheory as AbstractTransferTheory,
)
from .contrast_transfer_theory import (
    ContrastTransferFunction as ContrastTransferFunction,
    ContrastTransferTheory as ContrastTransferTheory,
)



import jax
import numpy as np
import pytest

import cryojax.simulator as cs
from cryojax.image import crop_to_shape
from cryojax.io import read_array_with_spacing_from_mrc


jax.config.update("jax_enable_x64", True)


@pytest.mark.parametrize("model", ["noisy_model"])
def test_real_shape(model, request):
    """Make sure shapes are as expected in real space."""
    model = request.getfixturevalue(model)
    image = model.render()
    padded_image = model.render(postprocess=False)
    assert image.shape == model.instrument_config.shape
    assert padded_image.shape == model.instrument_config.padded_shape


@pytest.mark.parametrize("model", ["noisy_model"])
def test_fourier_shape(model, request):
    """Make sure shapes are as expected in fourier space."""
    model = request.getfixturevalue(model)
    image = model.render(get_real=False)
    padded_image = model.render(postprocess=False, get_real=False)
    assert image.shape == model.instrument_config.frequency_grid_in_pixels.shape[0:2]
    assert (
        padded_image.shape
        == model.instrument_config.padded_frequency_grid_in_pixels.shape[0:2]
    )


@pytest.mark.parametrize("shape", [(65, 65), (65, 64), (64, 65)])
def test_even_vs_odd_image_shape(shape, sample_mrc_path, pixel_size):
    control_shape = (64, 64)
    real_voxel_grid, voxel_size = read_array_with_spacing_from_mrc(sample_mrc_path)
    potential = cs.FourierVoxelGridPotential.from_real_voxel_grid(
        real_voxel_grid, voxel_size
    )
    assert control_shape == potential.fourier_voxel_grid.shape[0:2]
    pose = cs.EulerAnglePose()
    method = cs.FourierSliceExtraction()
    specimen = cs.SingleStructureEnsemble(potential, pose)
    transfer_theory = cs.ContrastTransferTheory(cs.ContrastTransferFunction())
    theory = cs.WeakPhaseScatteringTheory(specimen, method, transfer_theory)
    config_control = cs.InstrumentConfig(
        control_shape, pixel_size, voltage_in_kilovolts=300.0
    )
    config_test = cs.InstrumentConfig(shape, pixel_size, voltage_in_kilovolts=300.0)
    pipeline_control = cs.ContrastImagingPipeline(config_control, theory)
    pipeline_test = cs.ContrastImagingPipeline(config_test, theory)

    np.testing.assert_allclose(
        crop_to_shape(pipeline_test.render(), control_shape),
        pipeline_control.render(),
        atol=1e-4,
    )



from ._assembly import (
    AbstractAssembly as AbstractAssembly,
    compute_helical_lattice_positions as compute_helical_lattice_positions,
    compute_helical_lattice_rotations as compute_helical_lattice_rotations,
    HelicalAssembly as HelicalAssembly,
)
from ._detector import (
    AbstractDetector as AbstractDetector,
    AbstractDQE as AbstractDQE,
    GaussianDetector as GaussianDetector,
    IdealCountingDQE as IdealCountingDQE,
    IdealDQE as IdealDQE,
    PoissonDetector as PoissonDetector,
)
from ._imaging_pipeline import (
    AbstractImagingPipeline as AbstractImagingPipeline,
    ContrastImagingPipeline as ContrastImagingPipeline,
    ElectronCountingImagingPipeline as ElectronCountingImagingPipeline,
    IntensityImagingPipeline as IntensityImagingPipeline,
)
from ._instrument_config import InstrumentConfig as InstrumentConfig
from ._pose import (
    AbstractPose as AbstractPose,
    AxisAnglePose as AxisAnglePose,
    EulerAnglePose as EulerAnglePose,
    QuaternionPose as QuaternionPose,
)
from ._potential_integrator import (
    AbstractFourierVoxelExtraction as AbstractFourierVoxelExtraction,
    AbstractPotentialIntegrator as AbstractPotentialIntegrator,
    AbstractVoxelPotentialIntegrator as AbstractVoxelPotentialIntegrator,
    FourierSliceExtraction as FourierSliceExtraction,
    GaussianMixtureProjection as GaussianMixtureProjection,
    NufftProjection as NufftProjection,
)
from ._potential_representation import (
    AbstractAtomicPotential as AbstractAtomicPotential,
    AbstractPotentialRepresentation as AbstractPotentialRepresentation,
    AbstractTabulatedAtomicPotential as AbstractTabulatedAtomicPotential,
    AbstractVoxelPotential as AbstractVoxelPotential,
    FourierVoxelGridPotential as FourierVoxelGridPotential,
    FourierVoxelGridPotentialInterpolator as FourierVoxelGridPotentialInterpolator,
    GaussianMixtureAtomicPotential as GaussianMixtureAtomicPotential,
    PengAtomicPotential as PengAtomicPotential,
    RealVoxelCloudPotential as RealVoxelCloudPotential,
    RealVoxelGridPotential as RealVoxelGridPotential,
)
from ._scattering_theory import (
    AbstractScatteringTheory as AbstractScatteringTheory,
    AbstractWeakPhaseScatteringTheory as AbstractWeakPhaseScatteringTheory,
    compute_phase_shifts_from_integrated_potential as compute_phase_shifts_from_integrated_potential,  # noqa: E501
    LinearSuperpositionScatteringTheory as LinearSuperpositionScatteringTheory,
    WeakPhaseScatteringTheory as WeakPhaseScatteringTheory,
)
from ._solvent import (
    AbstractIce as AbstractIce,
    GaussianIce as GaussianIce,
)
from ._structural_ensemble import (
    AbstractConformationalVariable as AbstractConformationalVariable,
    AbstractStructuralEnsemble as AbstractStructuralEnsemble,
    AbstractStructuralEnsembleBatcher as AbstractStructuralEnsembleBatcher,
    DiscreteConformationalVariable as DiscreteConformationalVariable,
    DiscreteStructuralEnsemble as DiscreteStructuralEnsemble,
    SingleStructureEnsemble as SingleStructureEnsemble,
)
from ._transfer_theory import (
    AbstractTransferFunction as AbstractTransferFunction,
    AbstractTransferTheory as AbstractTransferTheory,
    ContrastTransferFunction as ContrastTransferFunction,
    ContrastTransferTheory as ContrastTransferTheory,
)



"""Functions common to multiple scattering theories."""

import jax.numpy as jnp
from jaxtyping import Array, Float, Inexact


def compute_phase_shifts_from_integrated_potential(
    integrated_potential: Inexact[Array, "y_dim x_dim"],
    wavelength_in_angstroms: Float[Array, ""] | float,
) -> Inexact[Array, "y_dim x_dim"]:
    """Given an integrated potential, compute a phase shift distribution.

    !!! info
        In the projection approximation in cryo-EM, the phase shifts in the
        exit plane are given by

        $$\\eta(x, y) = \\sigma \\int dz \\ V(x, y, z),$$

        where $\\sigma$ is typically referred to as the interaction
        constant. However, in `cryojax`, the potential is rescaled
        to units of inverse length squared as

        $$U(x, y, z) = \\frac{2 m e}{\\hbar^2} V(x, y, z).$$

        With this rescaling of the potential, the phase shifts are equal to

        $$\\eta(x, y) = \\frac{\\lambda}{4 \\pi} \\int dz \\ U(x, y, z).$$

        **References**:

        - For the definition of the rescaled potential, see
        Chapter 69, Page 2003, Equation 69.6 from *Hawkes, Peter W., and Erwin Kasper.
        Principles of Electron Optics, Volume 4: Advanced Wave Optics. Academic Press,
        2022.*
        - For the definition of the phase shifts in terms of the rescaled potential, see
        Chapter 69, Page 2012, Equation 69.34b from *Hawkes, Peter W., and Erwin Kasper.
        Principles of Electron Optics, Volume 4: Advanced Wave Optics. Academic Press,
        2022.*

    See the documentation on atom-based scattering potentials for more information.
    """  # noqa: E501
    return integrated_potential * jnp.asarray(wavelength_in_angstroms) / (4 * jnp.pi)



"""
Routines for loading atomic structures.
Large amounts of the code are adapted from the ioSPI package
"""

import importlib.resources as pkg_resources
import os

import jax.numpy as jnp
import numpy as np
from jaxtyping import Array, Float, Int


def _load_peng_element_scattering_factor_parameter_table():
    """Internal function to load the atomic scattering factor parameter
    table."""
    with pkg_resources.as_file(
        pkg_resources.files("cryojax").joinpath("constants")
    ) as path:
        atom_scattering_factor_params = jnp.load(
            os.path.join(path, "peng1996_element_params.npy")
        )

    return jnp.asarray(atom_scattering_factor_params)


peng_element_scattering_factor_parameter_table = (
    _load_peng_element_scattering_factor_parameter_table()
)


def get_tabulated_scattering_factor_parameters(
    atom_identities: Int[Array, " n_atoms"] | Int[np.ndarray, " n_atoms"],
    scattering_factor_parameter_table: (
        Float[Array, "n_params n_elements n_scattering_factors"]
        | Float[np.ndarray, "n_params n_elements n_scattering_factors"]
    ) = peng_element_scattering_factor_parameter_table,
) -> Float[Array, "n_params n_atoms n_scattering_factors"]:
    """Gets the parameters for the scattering factor for each atom in
    `atom_names`.

    **Arguments:**

    - `atom_identitites`:
        Array containing the index of the one-hot encoded atom names.
        By default, Hydrogen is "1", Carbon is "6", Nitrogen is "7", etc.
    - `scattering_factor_parameter_table`:
        Array containing the table of scattering factors. By default, this
        is the tabulation from "Robust Parameterization of Elastic and
        Absorptive Electron Atomic Scattering Factors" by Peng et al. (1996).

    **Returns:**

    The particular scattering factor parameters stored in
    `scattering_factor_parameter_table` for `atom_identities`.
    """
    return jnp.asarray(scattering_factor_parameter_table)[:, jnp.asarray(atom_identities)]



import jax.numpy as jnp
from jaxtyping import Array, Float

import cryojax.simulator as cs


def test_voxel_electron_potential_loaders():
    real_voxel_grid = jnp.zeros((10, 10, 10), dtype=float)
    voxel_size = 1.1
    fourier_potential = cs.FourierVoxelGridPotential.from_real_voxel_grid(
        real_voxel_grid, voxel_size=voxel_size
    )
    real_potential = cs.RealVoxelGridPotential.from_real_voxel_grid(
        real_voxel_grid, voxel_size=voxel_size
    )
    cloud_potential = cs.RealVoxelCloudPotential.from_real_voxel_grid(
        real_voxel_grid, voxel_size=voxel_size
    )
    for potential in [real_potential, fourier_potential, cloud_potential]:
        assert potential.voxel_size == jnp.asarray(voxel_size)

    assert isinstance(
        fourier_potential.frequency_slice_in_pixels,
        Float[Array, "1 _ _ 3"],
    )
    assert isinstance(real_potential.coordinate_grid_in_pixels, Float[Array, "_ _ _ 3"])
    assert isinstance(cloud_potential.coordinate_list_in_pixels, Float[Array, "_ 3"])



from . import distributions as distributions
from ._grid_search import (
    AbstractGridSearchMethod as AbstractGridSearchMethod,
    MinimumSearchMethod as MinimumSearchMethod,
    MinimumSolution as MinimumSolution,
    MinimumState as MinimumState,
    PyTreeGrid as PyTreeGrid,
    PyTreeGridIndex as PyTreeGridIndex,
    PyTreeGridPoint as PyTreeGridPoint,
    run_grid_search as run_grid_search,
    tree_grid_shape as tree_grid_shape,
    tree_grid_take as tree_grid_take,
    tree_grid_unravel_index as tree_grid_unravel_index,
)
from ._transforms import (
    AbstractLieGroupTransform as AbstractLieGroupTransform,
    AbstractParameterTransform as AbstractParameterTransform,
    apply_updates_with_lie_transform as apply_updates_with_lie_transform,
    ComposedTransform as ComposedTransform,
    ExpTransform as ExpTransform,
    RescalingTransform as RescalingTransform,
    resolve_transforms as resolve_transforms,
    SE3Transform as SE3Transform,
    SO3Transform as SO3Transform,
)



from ._scattering_factor_parameters import (
    get_tabulated_scattering_factor_parameters as get_tabulated_scattering_factor_parameters,  # noqa: E501
    peng_element_scattering_factor_parameter_table as peng_element_scattering_factor_parameter_table,  # noqa: E501
)
from ._unit_conversions import (
    convert_keV_to_angstroms as convert_keV_to_angstroms,
)



from .assembly import AbstractAssembly as AbstractAssembly
from .helix import (
    compute_helical_lattice_positions as compute_helical_lattice_positions,
    compute_helical_lattice_rotations as compute_helical_lattice_rotations,
    HelicalAssembly as HelicalAssembly,
)



import jax.numpy as jnp
from jaxtyping import Array, Float

from ...coordinates import cartesian_to_polar


# Not currently public API
def compute_phase_shifts(
    frequency_grid_in_angstroms: Float[Array, "y_dim x_dim 2"],
    defocus_axis_1_in_angstroms: Float[Array, ""],
    defocus_axis_2_in_angstroms: Float[Array, ""],
    astigmatism_angle: Float[Array, ""],
    wavelength_in_angstroms: Float[Array, ""],
    spherical_aberration_in_angstroms: Float[Array, ""],
    phase_shift: Float[Array, ""],
) -> Float[Array, "y_dim x_dim"]:
    k_sqr, azimuth = cartesian_to_polar(frequency_grid_in_angstroms, square=True)
    defocus = 0.5 * (
        defocus_axis_1_in_angstroms
        + defocus_axis_2_in_angstroms
        + (defocus_axis_1_in_angstroms - defocus_axis_2_in_angstroms)
        * jnp.cos(2.0 * (azimuth - astigmatism_angle))
    )
    defocus_phase_shifts = -0.5 * defocus * wavelength_in_angstroms * k_sqr
    aberration_phase_shifts = (
        0.25
        * spherical_aberration_in_angstroms
        * (wavelength_in_angstroms**3)
        * (k_sqr**2)
    )
    phase_shifts = (2 * jnp.pi) * (
        defocus_phase_shifts + aberration_phase_shifts
    ) - phase_shift

    return phase_shifts


# Not currently public API
def compute_phase_shifts_with_amplitude_contrast_ratio(
    frequency_grid_in_angstroms: Float[Array, "y_dim x_dim 2"],
    defocus_axis_1_in_angstroms: Float[Array, ""],
    defocus_axis_2_in_angstroms: Float[Array, ""],
    astigmatism_angle: Float[Array, ""],
    wavelength_in_angstroms: Float[Array, ""],
    spherical_aberration_in_angstroms: Float[Array, ""],
    phase_shift: Float[Array, ""],
    amplitude_contrast_ratio: Float[Array, ""],
) -> Float[Array, "y_dim x_dim"]:
    phase_shifts = compute_phase_shifts(
        frequency_grid_in_angstroms,
        defocus_axis_1_in_angstroms,
        defocus_axis_2_in_angstroms,
        astigmatism_angle,
        wavelength_in_angstroms,
        spherical_aberration_in_angstroms,
        phase_shift,
    )
    amplitude_contrast_phase_shifts = jnp.arctan(
        amplitude_contrast_ratio / jnp.sqrt(1.0 - amplitude_contrast_ratio**2)
    )
    phase_shifts -= amplitude_contrast_phase_shifts

    return phase_shifts



"""
Image formation models.
"""

from abc import abstractmethod
from typing import Optional
from typing_extensions import override

import jax
from equinox import AbstractVar, Module
from jaxtyping import Array, Complex, Float, PRNGKeyArray

from ..image import irfftn, rfftn
from ..image.operators import FilterLike, MaskLike
from ._detector import AbstractDetector
from ._instrument_config import InstrumentConfig
from ._scattering_theory import AbstractScatteringTheory


class AbstractImagingPipeline(Module, strict=True):
    """Base class for an image formation model.

    Call an `AbstractImagingPipeline`'s `render` routine.
    """

    instrument_config: AbstractVar[InstrumentConfig]
    filter: AbstractVar[Optional[FilterLike]]
    mask: AbstractVar[Optional[MaskLike]]

    @abstractmethod
    def render(
        self,
        rng_key: Optional[PRNGKeyArray] = None,
        *,
        postprocess: bool = True,
        get_real: bool = True,
    ) -> (
        Float[Array, "{self.instrument_config.y_dim} {self.instrument_config.x_dim}"]
        | Float[
            Array,
            "{self.instrument_config.padded_y_dim} "
            "{self.instrument_config.padded_x_dim}",
        ]
        | Complex[
            Array, "{self.instrument_config.y_dim} {self.instrument_config.x_dim//2+1}"
        ]
        | Complex[
            Array,
            "{self.instrument_config.padded_y_dim} "
            "{self.instrument_config.padded_x_dim//2+1}",
        ]
    ):
        """Render an image without any stochasticity.

        **Arguments:**

        - `rng_key`: The random number generator key. If not passed, render an image
                     with no stochasticity.
        - `postprocess`: If `True`, view the cropped, filtered, and masked image.
                          If `postprocess = False`, `ImagePipeline.filter`,
                          `ImagePipeline.mask`, and cropping to `InstrumentConfig.shape`
                          are not applied. Instead, an image at the shape
                          `Instrument.padded_shape` is returned.
        - `get_real`: If `True`, return the image in real space.
        """
        raise NotImplementedError

    def postprocess(
        self,
        image: Complex[
            Array,
            "{self.instrument_config.padded_y_dim} "
            "{self.instrument_config.padded_x_dim//2+1}",
        ],
        *,
        get_real: bool = True,
    ) -> (
        Float[Array, "{self.instrument_config.y_dim} {self.instrument_config.x_dim}"]
        | Complex[
            Array,
            "{self.instrument_config.y_dim} {self.instrument_config.x_dim//2+1}",
        ]
    ):
        """Return an image postprocessed with filters, cropping, and masking
        in either real or fourier space.
        """
        instrument_config = self.instrument_config
        if (
            self.mask is None
            and instrument_config.padded_shape == instrument_config.shape
        ):
            # ... if there are no masks and we don't need to crop,
            # minimize moving back and forth between real and fourier space
            if self.filter is not None:
                image = self.filter(image)
            return irfftn(image, s=instrument_config.shape) if get_real else image
        else:
            # ... otherwise, apply filter, crop, and mask, again trying to
            # minimize moving back and forth between real and fourier space
            is_filter_applied = True if self.filter is None else False
            if (
                self.filter is not None
                and self.filter.array.shape
                == instrument_config.padded_frequency_grid_in_pixels.shape[0:2]
            ):
                # ... apply the filter here if it is the same size as the padded
                # coordinates
                is_filter_applied = True
                image = self.filter(image)
            image = irfftn(image, s=instrument_config.padded_shape)
            image = instrument_config.crop_to_shape(image)
            if self.mask is not None:
                image = self.mask(image)
            if is_filter_applied or self.filter is None:
                return image if get_real else rfftn(image)
            else:
                # ... otherwise, apply the filter here and return. assume
                # the filter is the same size as the non-padded coordinates
                image = self.filter(rfftn(image))
                return irfftn(image, s=instrument_config.shape) if get_real else image

    def _maybe_postprocess(
        self,
        image: Complex[
            Array,
            "{self.instrument_config.padded_y_dim} "
            "{self.instrument_config.padded_x_dim//2+1}",
        ],
        *,
        postprocess: bool = True,
        get_real: bool = True,
    ) -> (
        Float[Array, "{self.instrument_config.y_dim} {self.instrument_config.x_dim}"]
        | Float[
            Array,
            "{self.instrument_config.padded_y_dim} "
            "{self.instrument_config.padded_x_dim}",
        ]
        | Complex[
            Array, "{self.instrument_config.y_dim} {self.instrument_config.x_dim//2+1}"
        ]
        | Complex[
            Array,
            "{self.instrument_config.padded_y_dim} "
            "{self.instrument_config.padded_x_dim//2+1}",
        ]
    ):
        instrument_config = self.instrument_config
        if postprocess:
            return self.postprocess(image, get_real=get_real)
        else:
            return irfftn(image, s=instrument_config.padded_shape) if get_real else image


class ContrastImagingPipeline(AbstractImagingPipeline, strict=True):
    """An image formation pipeline that returns the image contrast from a linear
    scattering theory.

    **Attributes:**

    - `instrument_config`: The configuration of the instrument, such as for the pixel size
                and the wavelength.
    - `scattering_theory`: The scattering theory. This must be a linear scattering
                           theory.
    - `filter: `A filter to apply to the image.
    - `mask`: A mask to apply to the image.
    """

    instrument_config: InstrumentConfig
    scattering_theory: AbstractScatteringTheory

    filter: Optional[FilterLike]
    mask: Optional[MaskLike]

    def __init__(
        self,
        instrument_config: InstrumentConfig,
        scattering_theory: AbstractScatteringTheory,
        *,
        filter: Optional[FilterLike] = None,
        mask: Optional[MaskLike] = None,
    ):
        self.instrument_config = instrument_config
        self.scattering_theory = scattering_theory
        self.filter = filter
        self.mask = mask

    @override
    def render(
        self,
        rng_key: Optional[PRNGKeyArray] = None,
        *,
        postprocess: bool = True,
        get_real: bool = True,
    ) -> (
        Float[Array, "{self.instrument_config.y_dim} {self.instrument_config.x_dim}"]
        | Float[
            Array,
            "{self.instrument_config.padded_y_dim} "
            "{self.instrument_config.padded_x_dim}",
        ]
        | Complex[
            Array, "{self.instrument_config.y_dim} {self.instrument_config.x_dim//2+1}"
        ]
        | Complex[
            Array,
            "{self.instrument_config.padded_y_dim} "
            "{self.instrument_config.padded_x_dim//2+1}",
        ]
    ):
        # Compute the squared wavefunction
        fourier_contrast_at_detector_plane = (
            self.scattering_theory.compute_fourier_contrast_at_detector_plane(
                self.instrument_config, rng_key
            )
        )

        return self._maybe_postprocess(
            fourier_contrast_at_detector_plane, postprocess=postprocess, get_real=get_real
        )


class IntensityImagingPipeline(AbstractImagingPipeline, strict=True):
    """An image formation pipeline that returns an intensity distribution---or in other
    words a squared wavefunction.

    **Attributes:**

    - `instrument_config`: The configuration of the instrument, such as for the pixel size
                and the wavelength.
    - `scattering_theory`: The scattering theory.
    - `filter: `A filter to apply to the image.
    - `mask`: A mask to apply to the image.
    """

    instrument_config: InstrumentConfig
    scattering_theory: AbstractScatteringTheory

    filter: Optional[FilterLike]
    mask: Optional[MaskLike]

    def __init__(
        self,
        instrument_config: InstrumentConfig,
        scattering_theory: AbstractScatteringTheory,
        *,
        filter: Optional[FilterLike] = None,
        mask: Optional[MaskLike] = None,
    ):
        self.instrument_config = instrument_config
        self.scattering_theory = scattering_theory
        self.filter = filter
        self.mask = mask

    @override
    def render(
        self,
        rng_key: Optional[PRNGKeyArray] = None,
        *,
        postprocess: bool = True,
        get_real: bool = True,
    ) -> (
        Float[Array, "{self.instrument_config.y_dim} {self.instrument_config.x_dim}"]
        | Float[
            Array,
            "{self.instrument_config.padded_y_dim} "
            "{self.instrument_config.padded_x_dim}",
        ]
        | Complex[
            Array,
            "{self.instrument_config.y_dim} {self.instrument_config.x_dim//2+1}",
        ]
        | Complex[
            Array,
            "{self.instrument_config.padded_y_dim} "
            "{self.instrument_config.padded_x_dim//2+1}",
        ]
    ):
        theory = self.scattering_theory
        fourier_squared_wavefunction_at_detector_plane = (
            theory.compute_fourier_squared_wavefunction_at_detector_plane(
                self.instrument_config, rng_key
            )
        )

        return self._maybe_postprocess(
            fourier_squared_wavefunction_at_detector_plane,
            postprocess=postprocess,
            get_real=get_real,
        )


class ElectronCountingImagingPipeline(AbstractImagingPipeline, strict=True):
    """An image formation pipeline that returns electron counts, given a
    model for the detector.

    **Attributes:**

    - `instrument_config`: The configuration of the instrument, such as for the pixel size
                and the wavelength.
    - `scattering_theory`: The scattering theory.
    - `detector`: The electron detector.
    - `filter: `A filter to apply to the image.
    - `mask`: A mask to apply to the image.
    """

    instrument_config: InstrumentConfig
    scattering_theory: AbstractScatteringTheory
    detector: AbstractDetector

    filter: Optional[FilterLike]
    mask: Optional[MaskLike]

    def __init__(
        self,
        instrument_config: InstrumentConfig,
        scattering_theory: AbstractScatteringTheory,
        detector: AbstractDetector,
        *,
        filter: Optional[FilterLike] = None,
        mask: Optional[MaskLike] = None,
    ):
        self.instrument_config = instrument_config
        self.scattering_theory = scattering_theory
        self.detector = detector
        self.filter = filter
        self.mask = mask

    @override
    def render(
        self,
        rng_key: Optional[PRNGKeyArray] = None,
        *,
        postprocess: bool = True,
        get_real: bool = True,
    ) -> (
        Float[Array, "{self.instrument_config.y_dim} {self.instrument_config.x_dim}"]
        | Float[
            Array,
            "{self.instrument_config.padded_y_dim} "
            "{self.instrument_config.padded_x_dim}",
        ]
        | Complex[
            Array, "{self.instrument_config.y_dim} {self.instrument_config.x_dim//2+1}"
        ]
        | Complex[
            Array,
            "{self.instrument_config.padded_y_dim} "
            "{self.instrument_config.padded_x_dim//2+1}",
        ]
    ):
        if rng_key is None:
            # Compute the squared wavefunction
            theory = self.scattering_theory
            fourier_squared_wavefunction_at_detector_plane = (
                theory.compute_fourier_squared_wavefunction_at_detector_plane(
                    self.instrument_config
                )
            )
            # ... now measure the expected electron events at the detector
            fourier_expected_electron_events = (
                self.detector.compute_expected_electron_events(
                    fourier_squared_wavefunction_at_detector_plane, self.instrument_config
                )
            )

            return self._maybe_postprocess(
                fourier_expected_electron_events,
                postprocess=postprocess,
                get_real=get_real,
            )
        else:
            keys = jax.random.split(rng_key)
            # Compute the squared wavefunction
            theory = self.scattering_theory
            fourier_squared_wavefunction_at_detector_plane = (
                theory.compute_fourier_squared_wavefunction_at_detector_plane(
                    self.instrument_config, keys[0]
                )
            )
            # ... now measure the detector readout
            fourier_detector_readout = self.detector.compute_detector_readout(
                keys[1],
                fourier_squared_wavefunction_at_detector_plane,
                self.instrument_config,
            )

            return self._maybe_postprocess(
                fourier_detector_readout,
                postprocess=postprocess,
                get_real=get_real,
            )



from abc import abstractmethod
from typing import Optional

from equinox import Module
from jaxtyping import Array, Complex, Float

from ...image.operators import (
    AbstractFourierOperator,
)
from .._instrument_config import InstrumentConfig


class AbstractTransferFunction(AbstractFourierOperator, strict=True):
    """An abstract base class for a transfer function."""

    @abstractmethod
    def __call__(
        self,
        frequency_grid_in_angstroms: Float[Array, "y_dim x_dim 2"],
        *,
        wavelength_in_angstroms: Optional[Float[Array, ""] | float] = None,
        defocus_offset: Float[Array, ""] | float = 0.0,
    ) -> Float[Array, "y_dim x_dim"] | Complex[Array, "y_dim x_dim"]:
        raise NotImplementedError


class AbstractTransferTheory(Module, strict=True):
    """Base class for a transfer theory."""

    @abstractmethod
    def __call__(
        self,
        fourier_phase_or_wavefunction_at_exit_plane: (
            Complex[
                Array,
                "{instrument_config.padded_y_dim} "
                "{instrument_config.padded_x_dim//2+1}",
            ]
            | Complex[
                Array,
                "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim}",
            ]
        ),
        instrument_config: InstrumentConfig,
        defocus_offset: Float[Array, ""] | float = 0.0,
    ) -> (
        Complex[
            Array,
            "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}",
        ]
        | Complex[
            Array, "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim}"
        ]
    ):
        """Pass an image through the transfer theory."""
        raise NotImplementedError



"""
Abstraction of electron detectors in a cryo-EM image.
"""

from abc import abstractmethod
from typing import Optional
from typing_extensions import override

import jax.numpy as jnp
import jax.random as jr
import numpy as np
from equinox import AbstractVar, field, Module
from jaxtyping import Array, Complex, Float, PRNGKeyArray

from .._errors import error_if_not_fractional
from ..image import irfftn, rfftn
from ..image.operators import AbstractFourierOperator
from ._instrument_config import InstrumentConfig


class AbstractDQE(AbstractFourierOperator, strict=True):
    r"""Base class for a detector DQE."""

    fraction_detected_electrons: AbstractVar[Float[Array, ""]]

    @abstractmethod
    def __call__(
        self,
        frequency_grid_in_angstroms_or_pixels: Float[Array, "y_dim x_dim 2"],
        *,
        pixel_size: Optional[Float[Array, ""]] = None,
    ) -> Float[Array, "y_dim x_dim"]:
        """**Arguments:**

        - `frequency_grid_in_angstroms_or_pixels`: A frequency grid
                                                   given in angstroms
                                                   or pixels. If given
                                                   in angstroms, `pixel_size`
                                                   must be passed
        - `pixel_size`: The pixel size of `frequency_grid_in_angstroms_or_pixels`.
        """
        raise NotImplementedError


class IdealCountingDQE(AbstractDQE, strict=True):
    r"""A perfect DQE for a detector at a discrete pixel size.

    See Ruskin et. al. "Quantitative characterization of electron detectors for
    transmission electron microscopy." (2013) for details.
    """

    fraction_detected_electrons: Float[Array, ""] = field(
        default=1.0, converter=error_if_not_fractional
    )

    @override
    def __call__(
        self,
        frequency_grid_in_angstroms_or_pixels: Float[Array, "y_dim x_dim 2"],
        *,
        pixel_size: Optional[Float[Array, ""]] = None,
    ) -> Float[Array, "y_dim x_dim"]:
        if pixel_size is None:
            frequency_grid_in_nyquist_units = frequency_grid_in_angstroms_or_pixels / 0.5
        else:
            frequency_grid_in_nyquist_units = (
                frequency_grid_in_angstroms_or_pixels * pixel_size
            ) / 0.5
        return (
            self.fraction_detected_electrons**2
            * jnp.sinc(frequency_grid_in_nyquist_units[..., 0] / 2) ** 2
            * jnp.sinc(frequency_grid_in_nyquist_units[..., 1] / 2) ** 2
        )


class IdealDQE(AbstractDQE, strict=True):
    r"""A DQE that is perfect across all spatial frequencies."""

    fraction_detected_electrons: Float[Array, ""] = field(
        default=1.0, converter=error_if_not_fractional
    )

    @override
    def __call__(
        self,
        frequency_grid_in_angstroms_or_pixels: Float[Array, "y_dim x_dim 2"],
        *,
        pixel_size: Optional[Float[Array, ""]] = None,
    ) -> Float[Array, "y_dim x_dim"]:
        return jnp.full(
            frequency_grid_in_angstroms_or_pixels.shape[0:2],
            self.fraction_detected_electrons,
        )


class AbstractDetector(Module, strict=True):
    """Base class for an electron detector."""

    dqe: AbstractDQE

    def __init__(self, dqe: AbstractDQE):
        self.dqe = dqe

    @abstractmethod
    def sample_readout_from_expected_events(
        self, key: PRNGKeyArray, expected_electron_events: Float[Array, "y_dim x_dim"]
    ) -> Float[Array, "y_dim x_dim"]:
        """Sample a realization from the detector noise model."""
        raise NotImplementedError

    def compute_expected_electron_events(
        self,
        fourier_squared_wavefunction_at_detector_plane: Complex[
            Array,
            "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}",
        ],
        instrument_config: InstrumentConfig,
    ) -> Complex[
        Array, "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}"
    ]:
        """Compute the expected electron events from the detector."""
        fourier_expected_electron_events = (
            self._compute_expected_events_or_detector_readout(
                fourier_squared_wavefunction_at_detector_plane,
                instrument_config,
                key=None,
            )
        )

        return fourier_expected_electron_events

    def compute_detector_readout(
        self,
        key: PRNGKeyArray,
        fourier_squared_wavefunction_at_detector_plane: Complex[
            Array,
            "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}",
        ],
        instrument_config: InstrumentConfig,
    ) -> Complex[
        Array, "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}"
    ]:
        """Measure the readout from the detector."""
        fourier_detector_readout = self._compute_expected_events_or_detector_readout(
            fourier_squared_wavefunction_at_detector_plane,
            instrument_config,
            key,
        )

        return fourier_detector_readout

    def _compute_expected_events_or_detector_readout(
        self,
        fourier_squared_wavefunction_at_detector_plane: Complex[
            Array,
            "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}",
        ],
        instrument_config: InstrumentConfig,
        key: Optional[PRNGKeyArray] = None,
    ) -> Complex[
        Array, "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}"
    ]:
        """Pass the image through the detector model."""
        N_pix = np.prod(instrument_config.padded_shape)
        frequency_grid = instrument_config.padded_frequency_grid_in_pixels
        # Compute the time-integrated electron flux in pixels
        electrons_per_pixel = (
            instrument_config.electrons_per_angstrom_squared
            * instrument_config.pixel_size**2
        )
        # ... now the total number of electrons over the entire image
        electrons_per_image = N_pix * electrons_per_pixel
        # Normalize the squared wavefunction to a set of probabilities
        fourier_squared_wavefunction_at_detector_plane /= (
            fourier_squared_wavefunction_at_detector_plane[0, 0]
        )
        # Compute the noiseless signal by applying the DQE to the squared wavefunction
        fourier_signal = fourier_squared_wavefunction_at_detector_plane * jnp.sqrt(
            self.dqe(frequency_grid)
        )
        # Apply the integrated dose rate
        fourier_expected_electron_events = electrons_per_image * fourier_signal
        if key is None:
            # If there is no key given, return
            return fourier_expected_electron_events
        else:
            # ... otherwise, go to real space, sample, go back to fourier,
            # and return.
            expected_electron_events = irfftn(
                fourier_expected_electron_events, s=instrument_config.padded_shape
            )
            return rfftn(
                self.sample_readout_from_expected_events(key, expected_electron_events)
            )


class GaussianDetector(AbstractDetector, strict=True):
    """A detector with a gaussian noise model. This is the gaussian limit
    of `PoissonDetector`.
    """

    @override
    def sample_readout_from_expected_events(
        self, key: PRNGKeyArray, expected_electron_events: Float[Array, "y_dim x_dim"]
    ) -> Float[Array, "y_dim x_dim"]:
        return expected_electron_events + jnp.sqrt(expected_electron_events) * jr.normal(
            key, expected_electron_events.shape
        )


class PoissonDetector(AbstractDetector, strict=True):
    """A detector with a poisson noise model."""

    @override
    def sample_readout_from_expected_events(
        self, key: PRNGKeyArray, expected_electron_events: Float[Array, "y_dim x_dim"]
    ) -> Float[Array, "y_dim x_dim"]:
        return jr.poisson(key, expected_electron_events).astype(float)



from .base_conformation import (
    AbstractConformationalVariable as AbstractConformationalVariable,
)
from .base_ensemble import (
    AbstractStructuralEnsemble as AbstractStructuralEnsemble,
    SingleStructureEnsemble as SingleStructureEnsemble,
)
from .discrete_ensemble import (
    DiscreteConformationalVariable as DiscreteConformationalVariable,
    DiscreteStructuralEnsemble as DiscreteStructuralEnsemble,
)
from .ensemble_batcher import (
    AbstractStructuralEnsembleBatcher as AbstractStructuralEnsembleBatcher,
)



from abc import abstractmethod
from typing import Optional
from typing_extensions import override

import equinox as eqx
import jax
import jax.numpy as jnp
from jaxtyping import Array, Complex, PRNGKeyArray

from .._instrument_config import InstrumentConfig
from .._pose import AbstractPose
from .._potential_integrator import AbstractPotentialIntegrator
from .._solvent import AbstractIce
from .._structural_ensemble import (
    AbstractConformationalVariable,
    AbstractStructuralEnsemble,
    AbstractStructuralEnsembleBatcher,
)
from .._transfer_theory import ContrastTransferTheory
from .base_scattering_theory import AbstractScatteringTheory
from .common_functions import compute_phase_shifts_from_integrated_potential


class AbstractWeakPhaseScatteringTheory(AbstractScatteringTheory, strict=True):
    """Base class for a scattering theory in linear image formation theory
    (the weak-phase approximation).
    """

    @abstractmethod
    def compute_fourier_phase_shifts_at_exit_plane(
        self,
        instrument_config: InstrumentConfig,
        rng_key: Optional[PRNGKeyArray] = None,
    ) -> Complex[
        Array, "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}"
    ]:
        raise NotImplementedError

    @override
    def compute_fourier_squared_wavefunction_at_detector_plane(
        self,
        instrument_config: InstrumentConfig,
        rng_key: Optional[PRNGKeyArray] = None,
    ) -> Complex[
        Array, "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}"
    ]:
        """Compute the squared wavefunction at the detector plane, given the
        contrast.
        """
        N1, N2 = instrument_config.padded_shape
        # ... compute the squared wavefunction directly from the image contrast
        # as |psi|^2 = 1 + 2C.
        fourier_contrast_at_detector_plane = (
            self.compute_fourier_contrast_at_detector_plane(instrument_config, rng_key)
        )
        fourier_squared_wavefunction_at_detector_plane = (
            (2 * fourier_contrast_at_detector_plane).at[0, 0].add(1.0 * N1 * N2)
        )
        return fourier_squared_wavefunction_at_detector_plane


class WeakPhaseScatteringTheory(AbstractWeakPhaseScatteringTheory, strict=True):
    """Base linear image formation theory."""

    structural_ensemble: AbstractStructuralEnsemble
    potential_integrator: AbstractPotentialIntegrator
    transfer_theory: ContrastTransferTheory
    solvent: Optional[AbstractIce] = None

    @override
    def compute_fourier_phase_shifts_at_exit_plane(
        self,
        instrument_config: InstrumentConfig,
        rng_key: Optional[PRNGKeyArray] = None,
    ) -> Complex[
        Array, "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}"
    ]:
        # Compute the phase shifts in the exit plane
        fourier_phase_shifts_at_exit_plane = (
            _compute_phase_shifts_from_integrated_potential(
                self.structural_ensemble, self.potential_integrator, instrument_config
            )
        )

        if rng_key is not None:
            # Get the potential of the specimen plus the ice
            if self.solvent is not None:
                fourier_phase_shifts_at_exit_plane = (
                    self.solvent.compute_fourier_phase_shifts_with_ice(
                        rng_key, fourier_phase_shifts_at_exit_plane, instrument_config
                    )
                )

        return fourier_phase_shifts_at_exit_plane

    @override
    def compute_fourier_contrast_at_detector_plane(
        self,
        instrument_config: InstrumentConfig,
        rng_key: Optional[PRNGKeyArray] = None,
    ) -> Complex[
        Array, "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}"
    ]:
        fourier_phase_shifts_at_exit_plane = (
            self.compute_fourier_phase_shifts_at_exit_plane(instrument_config, rng_key)
        )
        fourier_contrast_at_detector_plane = self.transfer_theory(
            fourier_phase_shifts_at_exit_plane,
            instrument_config,
            defocus_offset=self.structural_ensemble.pose.offset_z_in_angstroms,
        )

        return fourier_contrast_at_detector_plane


WeakPhaseScatteringTheory.__init__.__doc__ = """**Arguments:**

- `structural_ensemble`: The structural ensemble of scattering potentials.
- `potential_integrator`: The method for integrating the scattering potential.
- `transfer_theory`: The contrast transfer theory.
- `solvent`: The model for the solvent.
"""


class LinearSuperpositionScatteringTheory(AbstractWeakPhaseScatteringTheory, strict=True):
    """Compute the superposition of images of the structural ensemble batch returned by
    the `AbstractStructuralEnsembleBatcher`. This must operate in the weak phase
    approximation.
    """

    structural_ensemble_batcher: AbstractStructuralEnsembleBatcher
    potential_integrator: AbstractPotentialIntegrator
    transfer_theory: ContrastTransferTheory
    solvent: Optional[AbstractIce] = None

    @override
    def compute_fourier_phase_shifts_at_exit_plane(
        self,
        instrument_config: InstrumentConfig,
        rng_key: Optional[PRNGKeyArray] = None,
    ) -> Complex[
        Array, "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}"
    ]:
        def compute_image(ensemble_mapped, ensemble_no_mapped, instrument_config):
            ensemble = eqx.combine(ensemble_mapped, ensemble_no_mapped)
            fourier_phase_shifts_at_exit_plane = (
                _compute_phase_shifts_from_integrated_potential(
                    ensemble, self.potential_integrator, instrument_config
                )
            )
            return fourier_phase_shifts_at_exit_plane

        @eqx.filter_jit
        def compute_image_superposition(
            ensemble_mapped, ensemble_no_mapped, instrument_config
        ):
            return jnp.sum(
                jax.lax.map(
                    lambda x: compute_image(x, ensemble_no_mapped, instrument_config),
                    ensemble_mapped,
                ),
                axis=0,
            )

        # Get the batch
        ensemble_batch = (
            self.structural_ensemble_batcher.get_batched_structural_ensemble()
        )
        # Setup vmap over the pose and conformation
        is_mapped = lambda x: isinstance(
            x, (AbstractPose, AbstractConformationalVariable)
        )
        to_mapped = jax.tree_util.tree_map(is_mapped, ensemble_batch, is_leaf=is_mapped)
        mapped, no_mapped = eqx.partition(ensemble_batch, to_mapped)

        fourier_phase_shifts_at_exit_plane = compute_image_superposition(
            mapped, no_mapped, instrument_config
        )

        if rng_key is not None:
            # Get the potential of the specimen plus the ice
            if self.solvent is not None:
                fourier_phase_shifts_at_exit_plane = (
                    self.solvent.compute_fourier_phase_shifts_with_ice(
                        rng_key, fourier_phase_shifts_at_exit_plane, instrument_config
                    )
                )

        return fourier_phase_shifts_at_exit_plane

    @override
    def compute_fourier_contrast_at_detector_plane(
        self,
        instrument_config: InstrumentConfig,
        rng_key: Optional[PRNGKeyArray] = None,
    ) -> Complex[
        Array, "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}"
    ]:
        def compute_image(ensemble_mapped, ensemble_no_mapped, instrument_config):
            ensemble = eqx.combine(ensemble_mapped, ensemble_no_mapped)
            fourier_phase_shifts_at_exit_plane = (
                _compute_phase_shifts_from_integrated_potential(
                    ensemble, self.potential_integrator, instrument_config
                )
            )
            fourier_contrast_at_detector_plane = self.transfer_theory(
                fourier_phase_shifts_at_exit_plane, instrument_config
            )

            return fourier_contrast_at_detector_plane

        @eqx.filter_jit
        def compute_image_superposition(
            ensemble_mapped, ensemble_no_mapped, instrument_config
        ):
            return jnp.sum(
                jax.lax.map(
                    lambda x: compute_image(x, ensemble_no_mapped, instrument_config),
                    ensemble_mapped,
                ),
                axis=0,
            )

        # Get the batch
        ensemble_batch = (
            self.structural_ensemble_batcher.get_batched_structural_ensemble()
        )
        # Setup vmap over the pose and conformation
        is_mapped = lambda x: isinstance(
            x, (AbstractPose, AbstractConformationalVariable)
        )
        to_mapped = jax.tree_util.tree_map(is_mapped, ensemble_batch, is_leaf=is_mapped)
        mapped, no_mapped = eqx.partition(ensemble_batch, to_mapped)

        fourier_contrast_at_detector_plane = compute_image_superposition(
            mapped, no_mapped, instrument_config
        )

        if rng_key is not None:
            # Get the contrast from the ice and add to that of the image batch
            if self.solvent is not None:
                fourier_ice_contrast_at_detector_plane = self.transfer_theory(
                    self.solvent.sample_fourier_phase_shifts_from_ice(
                        rng_key, instrument_config
                    ),
                    instrument_config,
                )
                fourier_contrast_at_detector_plane += (
                    fourier_ice_contrast_at_detector_plane
                )

        return fourier_contrast_at_detector_plane


LinearSuperpositionScatteringTheory.__init__.__doc__ = """**Arguments:**

- `structural_ensemble_batcher`: The batcher that computes the states that over which to
                                 compute a superposition of images. Most commonly, this
                                 would be an `AbstractAssembly` concrete class.
- `potential_integrator`: The method for integrating the specimen potential.
- `transfer_theory`: The contrast transfer theory.
- `solvent`: The model for the solvent.
"""


def _compute_phase_shifts_from_integrated_potential(
    structural_ensemble, potential_integrator, instrument_config
):
    # Get potential in the lab frame
    potential = structural_ensemble.get_potential_in_lab_frame()
    # Compute the phase shifts in the exit plane
    fourier_integrated_potential = (
        potential_integrator.compute_fourier_integrated_potential(
            potential, instrument_config
        )
    )
    # Compute in-plane translation through fourier phase shifts
    translational_phase_shifts = structural_ensemble.pose.compute_shifts(
        instrument_config.padded_frequency_grid_in_angstroms
    )
    # Compute the phase shifts in exit plane and multiply by the translation.
    phase_shifts_in_exit_plane = compute_phase_shifts_from_integrated_potential(
        fourier_integrated_potential, instrument_config.wavelength_in_angstroms
    )
    return phase_shifts_in_exit_plane * translational_phase_shifts



"""
Helper routines to compute power spectra.
"""

from typing import Optional

import jax.numpy as jnp
from jaxtyping import Array, Complex, Float

from ._average import compute_binned_radial_average


def compute_radially_averaged_powerspectrum(
    fourier_image: Complex[Array, "y_dim x_dim"] | Complex[Array, "z_dim y_dim x_dim"],
    radial_frequency_grid: (
        Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]
    ),
    pixel_size: Float[Array, ""] | float = 1.0,
    *,
    minimum_frequency: Optional[Float[Array, ""] | float] = None,
    maximum_frequency: Optional[Float[Array, ""] | float] = None,
) -> tuple[Float[Array, " n_bins"], Float[Array, " n_bins"]]:
    """Compute the power spectrum of an image averaged on a set
    of radial bins.

    **Arguments:**

    `fourier_image`:
        An image in Fourier space.
    `radial_frequency_grid`:
        The radial frequency coordinate system of `fourier_image`.
    `pixel_size`:
        The pixel size of `radial_frequency_grid`.
    `minimum_frequency`:
        Minimum frequency bin. By default, `0.0`.
    `minimum_frequency`:
        Maximum frequency bin. By default, `1 / (2 * pixel_size)`.

    **Returns:**

    A tuple of the radially averaged power spectrum and the frequency bins
    over which it is computed.
    """
    # Compute squared amplitudes
    squared_fourier_amplitudes = (fourier_image * jnp.conjugate(fourier_image)).real
    # Compute bins
    q_min = 0.0 if minimum_frequency is None else minimum_frequency
    q_max = (
        jnp.sqrt(2) / (pixel_size * 2.0)
        if maximum_frequency is None
        else maximum_frequency
    )
    q_step = 1.0 / (pixel_size * max(*squared_fourier_amplitudes.shape))
    frequency_bins = jnp.linspace(q_min, q_max, 1 + int((q_max - q_min) / q_step))
    # Compute radially averaged power spectrum as a 1D profile
    radially_averaged_powerspectrum = compute_binned_radial_average(
        squared_fourier_amplitudes, radial_frequency_grid, frequency_bins
    )

    return radially_averaged_powerspectrum, frequency_bins



"""
Using the fourier slice theorem for computing volume projections.
"""

from abc import abstractmethod
from typing import Optional
from typing_extensions import override

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, Complex, Float

from ...image import (
    irfftn,
    map_coordinates,
    map_coordinates_with_cubic_spline,
    rfftn,
)
from .._instrument_config import InstrumentConfig
from .._potential_representation import (
    FourierVoxelGridPotential,
    FourierVoxelGridPotentialInterpolator,
)
from .base_potential_integrator import AbstractVoxelPotentialIntegrator


class AbstractFourierVoxelExtraction(
    AbstractVoxelPotentialIntegrator[
        FourierVoxelGridPotential | FourierVoxelGridPotentialInterpolator
    ],
    strict=True,
):
    """Integrate points to the exit plane by extracting a voxel surface
    from a 3D voxel grid.
    """

    interpolation_order: eqx.AbstractVar[int]
    interpolation_mode: eqx.AbstractVar[str]
    interpolation_cval: eqx.AbstractVar[complex]

    @abstractmethod
    def extract_voxels_from_spline_coefficients(
        self,
        spline_coefficients: Complex[Array, "dim+2 dim+2 dim+2"],
        frequency_slice_in_pixels: Float[Array, "1 dim dim 3"],
        voxel_size: Float[Array, ""],
        wavelength_in_angstroms: Float[Array, ""],
    ) -> Complex[Array, "dim dim//2+1"]:
        raise NotImplementedError

    @abstractmethod
    def extract_voxels_from_grid_points(
        self,
        fourier_voxel_grid: Complex[Array, "dim dim dim"],
        frequency_slice_in_pixels: Float[Array, "1 dim dim 3"],
        voxel_size: Float[Array, ""],
        wavelength_in_angstroms: Float[Array, ""],
    ) -> Complex[Array, "dim dim//2+1"]:
        raise NotImplementedError

    @override
    def compute_fourier_integrated_potential(
        self,
        potential: FourierVoxelGridPotential | FourierVoxelGridPotentialInterpolator,
        instrument_config: InstrumentConfig,
    ) -> Complex[
        Array, "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}"
    ]:
        """Compute the integrated scattering potential at the `InstrumentConfig` settings
        of a voxel-based representation in fourier-space, using fourier slice extraction.

        **Arguments:**

        - `potential`: The scattering potential representation.
        - `instrument_config`: The configuration of the resulting image.

        **Returns:**

        The extracted fourier voxels of the `potential`, at the
        `instrument_config.padded_shape` and the `instrument_config.pixel_size`.
        """
        frequency_slice = potential.frequency_slice_in_pixels
        N = frequency_slice.shape[1]
        if potential.shape != (N, N, N):
            raise AttributeError(
                "Only cubic boxes are supported for fourier slice extraction."
            )
        # Compute the fourier projection
        if isinstance(potential, FourierVoxelGridPotentialInterpolator):
            fourier_projection = self.extract_voxels_from_spline_coefficients(
                potential.coefficients,
                frequency_slice,
                potential.voxel_size,
                instrument_config.wavelength_in_angstroms,
            )
        elif isinstance(potential, FourierVoxelGridPotential):
            fourier_projection = self.extract_voxels_from_grid_points(
                potential.fourier_voxel_grid,
                frequency_slice,
                potential.voxel_size,
                instrument_config.wavelength_in_angstroms,
            )
        else:
            raise ValueError(
                "Supported types for `potential` are `FourierVoxelGridPotential` and "
                "`FourierVoxelGridPotentialInterpolator`."
            )

        # Resize the image to match the InstrumentConfig.padded_shape
        if instrument_config.padded_shape != (N, N):
            fourier_projection = rfftn(
                instrument_config.crop_or_pad_to_padded_shape(
                    irfftn(fourier_projection, s=(N, N))
                )
            )
        return self._convert_raw_image_to_integrated_potential(
            fourier_projection, potential, instrument_config
        )


class FourierSliceExtraction(AbstractFourierVoxelExtraction, strict=True):
    """Integrate points to the exit plane using the Fourier projection-slice theorem.

    This extracts slices using interpolation methods housed in
    `cryojax.image.map_coordinates` and `cryojax.image.map_coordinates_with_cubic_spline`.
    """

    pixel_rescaling_method: Optional[str]
    interpolation_order: int
    interpolation_mode: str
    interpolation_cval: complex

    def __init__(
        self,
        *,
        pixel_rescaling_method: Optional[str] = None,
        interpolation_order: int = 1,
        interpolation_mode: str = "fill",
        interpolation_cval: complex = 0.0 + 0.0j,
    ):
        """**Arguments:**

        - `pixel_rescaling_method`:
            Method for rescaling the final image to the `InstrumentConfig`
            pixel size. See `cryojax.image.rescale_pixel_size` for documentation.
        - `interpolation_order`:
            The interpolation order. This can be `0` (nearest-neighbor), `1`
            (linear), or `3` (cubic).
            Note that this argument is ignored when using this object with a
            `FourierVoxelGridInterpolator`.
        - `interpolation_mode`:
            Specify how to handle out of bounds indexing. See
            `cryojax.image.map_coordinates` for documentation.
        - `interpolation_cval`:
            Value for filling out-of-bounds indices. Used only when
            `interpolation_mode = "fill"`.
        """
        self.pixel_rescaling_method = pixel_rescaling_method
        self.interpolation_order = interpolation_order
        self.interpolation_mode = interpolation_mode
        self.interpolation_cval = interpolation_cval

    @override
    def extract_voxels_from_spline_coefficients(
        self,
        spline_coefficients: Complex[Array, "dim+2 dim+2 dim+2"],
        frequency_slice_in_pixels: Float[Array, "1 dim dim 3"],
        voxel_size: Float[Array, ""],
        wavelength_in_angstroms: Float[Array, ""],
    ) -> Complex[Array, "dim dim//2+1"]:
        """Extract a fourier slice using the interpolation defined by
        `spline_coefficients` at coordinates `frequency_slice_in_pixels`.

        **Arguments:**

        - `spline_coefficients`:
            Spline coefficients of the density grid in fourier space.
            The coefficients should be computed from a `fourier_voxel_grid`
            with the zero frequency component in the center. These are
            typically computed with the function
            `cryojax.image.compute_spline_coefficients`.
        - `frequency_slice_in_pixels`:
            Frequency central slice coordinate system. The zero
            frequency component should be in the center.
        - `voxel_size`:
            The voxel size of the `fourier_voxel_grid`. This argument is
            not used in the `FourierSliceExtraction` class.
        - `wavelength_in_angstroms`:
            The wavelength of the incident electron beam. This argument is
            not used in the `FourierSliceExtraction` class.

        **Returns:**

        The interpolated fourier slice at coordinates `frequency_slice_in_pixels`.
        """
        return _extract_slice_with_cubic_spline(
            spline_coefficients,
            frequency_slice_in_pixels,
            mode=self.interpolation_mode,
            cval=self.interpolation_cval,
        )

    @override
    def extract_voxels_from_grid_points(
        self,
        fourier_voxel_grid: Complex[Array, "dim dim dim"],
        frequency_slice_in_pixels: Float[Array, "1 dim dim 3"],
        voxel_size: Float[Array, ""],
        wavelength_in_angstroms: Float[Array, ""],
    ) -> Complex[Array, "dim dim//2+1"]:
        """Extract a fourier slice of the `fourier_voxel_grid` at coordinates
        `frequency_slice_in_pixels`.

        **Arguments:**

        - `fourier_voxel_grid`:
            Density grid in fourier space. The zero frequency component
            should be in the center.
        - `frequency_slice_in_pixels`:
            Frequency central slice coordinate system. The zero
            frequency component should be in the center.
        - `voxel_size`:
            The voxel size of the `fourier_voxel_grid`. This argument is
            not used in the `FourierSliceExtraction` class.
        - `wavelength_in_angstroms`:
            The wavelength of the incident electron beam. This argument is
            not used in the `FourierSliceExtraction` class.

        **Returns:**

        The interpolated fourier slice at coordinates `frequency_slice_in_pixels`.
        """
        return _extract_slice(
            fourier_voxel_grid,
            frequency_slice_in_pixels,
            interpolation_order=self.interpolation_order,
            mode=self.interpolation_mode,
            cval=self.interpolation_cval,
        )


class EwaldSphereExtraction(AbstractFourierVoxelExtraction, strict=True):
    """Integrate points to the exit plane by extracting a surface of the ewald
    sphere in fourier space.

    This extracts surfaces using interpolation methods housed in
    `cryojax.image.map_coordinates` and `cryojax.image.map_coordinates_with_cubic_spline`.
    """

    pixel_rescaling_method: Optional[str]
    interpolation_order: int
    interpolation_mode: str
    interpolation_cval: complex

    def __init__(
        self,
        *,
        pixel_rescaling_method: Optional[str] = None,
        interpolation_order: int = 1,
        interpolation_mode: str = "fill",
        interpolation_cval: complex = 0.0 + 0.0j,
    ):
        """**Arguments:**

        - `pixel_rescaling_method`:
            Method for rescaling the final image to the `InstrumentConfig`
            pixel size. See `cryojax.image.rescale_pixel_size` for documentation.
        - `interpolation_order`:
            The interpolation order. This can be `0` (nearest-neighbor), `1`
            (linear), or `3` (cubic).
            Note that this argument is ignored when using this object with a
            `FourierVoxelGridInterpolator`.
        - `interpolation_mode`:
            Specify how to handle out of bounds indexing. See
            `cryojax.image.map_coordinates` for documentation.
        - `interpolation_cval`:
            Value for filling out-of-bounds indices. Used only when
            `interpolation_mode = "fill"`.
        """
        self.pixel_rescaling_method = pixel_rescaling_method
        self.interpolation_order = interpolation_order
        self.interpolation_mode = interpolation_mode
        self.interpolation_cval = interpolation_cval

    @override
    def extract_voxels_from_spline_coefficients(
        self,
        spline_coefficients: Complex[Array, "dim+2 dim+2 dim+2"],
        frequency_slice_in_pixels: Float[Array, "1 dim dim 3"],
        voxel_size: Float[Array, ""],
        wavelength_in_angstroms: Float[Array, ""],
    ) -> Complex[Array, "dim dim//2+1"]:
        """Extract an ewald sphere surface of the `fourier_voxel_grid` at
        coordinates normal to `frequency_slice_in_pixels` at wavelength
        `wavelength_in_angstroms`.

        **Arguments:**

        - `spline_coefficients`:
            Spline coefficients of the density grid in fourier space.
            The coefficients should be computed from a `fourier_voxel_grid`
            with the zero frequency component in the center. These are
            typically computed with the function
            `cryojax.image.compute_spline_coefficients`.
        - `frequency_slice_in_pixels`:
            Frequency central slice coordinate system. The zero
            frequency component should be in the center.
        - `voxel_size`:
            The voxel size of the `fourier_voxel_grid`.
        - `wavelength_in_angstroms`:
            The wavelength of the incident electron beam.

        **Returns:**

        The interpolated ewald sphere surface at coordinates normal to
        `frequency_slice_in_pixels`.
        """
        return _extract_ewald_sphere_surface_with_cubic_spline(
            spline_coefficients,
            frequency_slice_in_pixels,
            voxel_size,
            wavelength_in_angstroms,
            mode=self.interpolation_mode,
            cval=self.interpolation_cval,
        )

    @override
    def extract_voxels_from_grid_points(
        self,
        fourier_voxel_grid: Complex[Array, "dim dim dim"],
        frequency_slice_in_pixels: Float[Array, "1 dim dim 3"],
        voxel_size: Float[Array, ""],
        wavelength_in_angstroms: Float[Array, ""],
    ) -> Complex[Array, "dim dim//2+1"]:
        """Extract an ewald sphere surface of the `fourier_voxel_grid` at
        coordinates normal to `frequency_slice_in_pixels` at wavelength
        `wavelength_in_angstroms`.

        **Arguments:**

        - `fourier_voxel_grid`:
            Density grid in fourier space. The zero frequency component
            should be in the center.
        - `frequency_slice_in_pixels`:
            Frequency central slice coordinate system. The zero
            frequency component should be in the center.
        - `voxel_size`:
            The voxel size of the `fourier_voxel_grid`.
        - `wavelength_in_angstroms`:
            The wavelength of the incident electron beam.

        **Returns:**

        The interpolated ewald sphere surface at coordinates normal to
        `frequency_slice_in_pixels`.
        """
        return _extract_ewald_sphere_surface(
            fourier_voxel_grid,
            frequency_slice_in_pixels,
            voxel_size,
            wavelength_in_angstroms,
            interpolation_order=self.interpolation_order,
            mode=self.interpolation_mode,
            cval=self.interpolation_cval,
        )


def _extract_slice(
    fourier_voxel_grid,
    frequency_slice,
    interpolation_order,
    **kwargs,
) -> Complex[Array, "dim dim//2+1"]:
    return _extract_surface_from_voxel_grid(
        fourier_voxel_grid,
        frequency_slice,
        is_spline_coefficients=False,
        interpolation_order=interpolation_order,
        **kwargs,
    )


def _extract_slice_with_cubic_spline(
    spline_coefficients, frequency_slice, **kwargs
) -> Complex[Array, "dim dim//2+1"]:
    return _extract_surface_from_voxel_grid(
        spline_coefficients, frequency_slice, is_spline_coefficients=True, **kwargs
    )


def _extract_ewald_sphere_surface(
    fourier_voxel_grid,
    frequency_slice,
    voxel_size,
    wavelength,
    interpolation_order,
    **kwargs,
) -> Complex[Array, "dim dim//2+1"]:
    ewald_sphere_frequencies = _get_ewald_sphere_surface_from_slice(
        frequency_slice, voxel_size, wavelength
    )
    return _extract_surface_from_voxel_grid(
        fourier_voxel_grid,
        ewald_sphere_frequencies,
        is_spline_coefficients=False,
        interpolation_order=interpolation_order,
        **kwargs,
    )


def _extract_ewald_sphere_surface_with_cubic_spline(
    spline_coefficients, frequency_slice, voxel_size, wavelength, **kwargs
) -> Complex[Array, "dim dim//2+1"]:
    ewald_sphere_frequencies = _get_ewald_sphere_surface_from_slice(
        frequency_slice, voxel_size, wavelength
    )
    return _extract_surface_from_voxel_grid(
        spline_coefficients,
        ewald_sphere_frequencies,
        is_spline_coefficients=True,
        **kwargs,
    )


def _get_ewald_sphere_surface_from_slice(
    frequency_slice_in_pixels: Float[Array, "1 dim dim 3"],
    voxel_size: Float[Array, ""],
    wavelength: Float[Array, ""],
) -> Float[Array, "1 dim dim 3"]:
    frequency_slice_with_zero_in_corner = jnp.fft.ifftshift(
        frequency_slice_in_pixels, axes=(0, 1, 2)
    )
    # Get zhat unit vector of the frequency slice
    xhat, yhat = (
        frequency_slice_with_zero_in_corner[0, 0, 1, :],
        frequency_slice_with_zero_in_corner[0, 1, 0, :],
    )
    xhat, yhat = xhat / jnp.linalg.norm(xhat), yhat / jnp.linalg.norm(yhat)
    zhat = jnp.cross(xhat, yhat)
    # Compute the ewald sphere surface, assuming the frequency slice is
    # in a rotated frame
    q_at_slice = frequency_slice_in_pixels
    q_squared = jnp.sum(q_at_slice**2, axis=-1)
    q_at_surface = (
        q_at_slice
        + (wavelength / voxel_size)
        * (q_squared[..., None] * zhat[None, None, None, :])
        / 2
    )
    return q_at_surface


def _extract_surface_from_voxel_grid(
    voxel_grid,
    frequency_coordinates,
    is_spline_coefficients=False,
    interpolation_order=1,
    **kwargs,
):
    # Convert to logical coordinates
    N = frequency_coordinates.shape[1]
    logical_frequency_slice = (frequency_coordinates * N) + N // 2
    # Convert arguments to map_coordinates convention and compute
    k_z, k_y, k_x = jnp.transpose(logical_frequency_slice, axes=[3, 0, 1, 2])
    if is_spline_coefficients:
        spline_coefficients = voxel_grid
        projection = map_coordinates_with_cubic_spline(
            spline_coefficients, (k_x, k_y, k_z), **kwargs
        )[0, :, :]
    else:
        fourier_voxel_grid = voxel_grid
        projection = map_coordinates(
            fourier_voxel_grid, (k_x, k_y, k_z), interpolation_order, **kwargs
        )[0, :, :]
    # Shift zero frequency component to corner and take upper half plane
    projection = jnp.fft.ifftshift(projection)[:, : N // 2 + 1]
    # Set last line of frequencies to zero if image dimension is even
    if N % 2 == 0:
        projection = projection.at[:, -1].set(0.0 + 0.0j).at[N // 2, :].set(0.0 + 0.0j)
    return projection



from ._filters import (
    AbstractFilter as AbstractFilter,
    CustomFilter as CustomFilter,
    FilterLike as FilterLike,
    HighpassFilter as HighpassFilter,
    InverseSincFilter as InverseSincFilter,
    LowpassFilter as LowpassFilter,
    WhiteningFilter as WhiteningFilter,
)
from ._fourier_operator import (
    AbstractFourierOperator as AbstractFourierOperator,
    FourierExp2D as FourierExp2D,
    FourierGaussian as FourierGaussian,
    FourierOperatorLike as FourierOperatorLike,
    Lorenzian as Lorenzian,
    ZeroMode as ZeroMode,
)
from ._masks import (
    AbstractMask as AbstractMask,
    CircularCosineMask as CircularCosineMask,
    CustomMask as CustomMask,
    Cylindrical2DCosineMask as Cylindrical2DCosineMask,
    MaskLike as MaskLike,
    SphericalCosineMask as SphericalCosineMask,
    SquareCosineMask as SquareCosineMask,
)
from ._operator import (
    AbstractImageMultiplier as AbstractImageMultiplier,
    AbstractImageOperator as AbstractImageOperator,
    Constant as Constant,
    DiffImageOperator as DiffImageOperator,
    Empirical as Empirical,
    Lambda as Lambda,
    ProductImageMultiplier as ProductImageMultiplier,
    ProductImageOperator as ProductImageOperator,
    SumImageOperator as SumImageOperator,
)
from ._real_operator import (
    AbstractRealOperator as AbstractRealOperator,
    Gaussian2D as Gaussian2D,
    RealOperatorLike as RealOperatorLike,
)



from typing import Optional
from typing_extensions import override

import jax.numpy as jnp
from equinox import field
from jaxtyping import Array, Complex, Float

from ..._errors import error_if_negative, error_if_not_fractional
from ...constants import convert_keV_to_angstroms
from ...image.operators import (
    Constant,
    FourierOperatorLike,
)
from .._instrument_config import InstrumentConfig
from .base_transfer_theory import AbstractTransferFunction, AbstractTransferTheory
from .common_functions import compute_phase_shifts_with_amplitude_contrast_ratio


class ContrastTransferFunction(AbstractTransferFunction, strict=True):
    """Compute an astigmatic Contrast Transfer Function (CTF) with a
    spherical aberration correction and amplitude contrast ratio.

    !!! info
        `cryojax` uses a convention different from CTFFIND for
        astigmatism parameters. It returns defocus major and minor
        axes, called "defocus1" and "defocus2". In order to convert
        from CTFFIND to `cryojax`,

        ```python
        defocus1, defocus2 = ... # Read from CTFFIND
        ctf = ContrastTransferFunction(
            defocus_in_angstroms=(defocus1+defocus2)/2,
            astigmatism_in_angstroms=defocus1-defocus2,
            ...
        )
        ```
    """

    defocus_in_angstroms: Float[Array, ""]
    astigmatism_in_angstroms: Float[Array, ""]
    astigmatism_angle: Float[Array, ""]
    voltage_in_kilovolts: Float[Array, ""] | float = field(static=True)
    spherical_aberration_in_mm: Float[Array, ""]
    amplitude_contrast_ratio: Float[Array, ""]
    phase_shift: Float[Array, ""]

    def __init__(
        self,
        defocus_in_angstroms: float | Float[Array, ""] = 10000.0,
        astigmatism_in_angstroms: float | Float[Array, ""] = 0.0,
        astigmatism_angle: float | Float[Array, ""] = 0.0,
        voltage_in_kilovolts: float | Float[Array, ""] = 300.0,
        spherical_aberration_in_mm: float | Float[Array, ""] = 2.7,
        amplitude_contrast_ratio: float | Float[Array, ""] = 0.1,
        phase_shift: float | Float[Array, ""] = 0.0,
    ):
        """**Arguments:**

        - `defocus_in_angstroms`: The mean defocus in Angstroms.
        - `astigmatism_in_angstroms`: The amount of astigmatism in Angstroms.
        - `astigmatism_angle`: The defocus angle.
        - `voltage_in_kilovolts`:
            The accelerating voltage in kV. This field is treated as *static*, i.e.
            as part of the pytree. This is because the accelerating voltage is treated
            as a traced value in the `InstrumentConfig`, since many modeling components
            are interested in the accelerating voltage.
        - `spherical_aberration_in_mm`: The spherical aberration coefficient in mm.
        - `amplitude_contrast_ratio`: The amplitude contrast ratio.
        - `phase_shift`: The additional phase shift.
        """
        self.defocus_in_angstroms = error_if_negative(defocus_in_angstroms)
        self.astigmatism_in_angstroms = jnp.asarray(astigmatism_in_angstroms)
        self.astigmatism_angle = jnp.asarray(astigmatism_angle)
        self.voltage_in_kilovolts = voltage_in_kilovolts
        self.spherical_aberration_in_mm = error_if_negative(spherical_aberration_in_mm)
        self.amplitude_contrast_ratio = error_if_not_fractional(amplitude_contrast_ratio)
        self.phase_shift = jnp.asarray(phase_shift)

    def __call__(
        self,
        frequency_grid_in_angstroms: Float[Array, "y_dim x_dim 2"],
        *,
        wavelength_in_angstroms: Optional[Float[Array, ""] | float] = None,
        defocus_offset: Float[Array, ""] | float = 0.0,
    ) -> Float[Array, "y_dim x_dim"]:
        # Convert degrees to radians
        phase_shift = jnp.deg2rad(self.phase_shift)
        astigmatism_angle = jnp.deg2rad(self.astigmatism_angle)
        # Convert spherical abberation coefficient to angstroms
        spherical_aberration_in_angstroms = self.spherical_aberration_in_mm * 1e7
        # Get the wavelength. It can either be passed from upstream or stored in the
        # CTF
        if wavelength_in_angstroms is None:
            wavelength_in_angstroms = convert_keV_to_angstroms(
                jnp.asarray(self.voltage_in_kilovolts)
            )
        else:
            wavelength_in_angstroms = jnp.asarray(wavelength_in_angstroms)
        defocus_axis_1_in_angstroms = (
            self.defocus_in_angstroms
            + jnp.asarray(defocus_offset)
            + self.astigmatism_in_angstroms / 2
        )
        defocus_axis_2_in_angstroms = (
            self.defocus_in_angstroms
            + jnp.asarray(defocus_offset)
            - self.astigmatism_in_angstroms / 2
        )
        # Compute phase shifts for CTF
        phase_shifts = compute_phase_shifts_with_amplitude_contrast_ratio(
            frequency_grid_in_angstroms,
            defocus_axis_1_in_angstroms,
            defocus_axis_2_in_angstroms,
            astigmatism_angle,
            wavelength_in_angstroms,
            spherical_aberration_in_angstroms,
            phase_shift,
            self.amplitude_contrast_ratio,
        )
        # Compute the CTF
        return jnp.sin(phase_shifts).at[0, 0].set(0.0)


class ContrastTransferTheory(AbstractTransferTheory, strict=True):
    """An optics model in the weak-phase approximation. Here, compute the image
    contrast by applying the CTF directly to the exit plane phase shifts.
    """

    ctf: ContrastTransferFunction
    envelope: FourierOperatorLike

    def __init__(
        self,
        ctf: ContrastTransferFunction,
        envelope: Optional[FourierOperatorLike] = None,
    ):
        """**Arguments:**

        - `ctf`: The contrast transfer function model.
        - `envelope`: The envelope function of the optics model.
        """

        self.ctf = ctf
        self.envelope = envelope or Constant(jnp.asarray(1.0))

    @override
    def __call__(
        self,
        fourier_phase_at_exit_plane: Complex[
            Array,
            "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}",
        ],
        instrument_config: InstrumentConfig,
        defocus_offset: Float[Array, ""] | float = 0.0,
    ) -> Complex[
        Array, "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}"
    ]:
        """Apply the CTF directly to the phase shifts in the exit plane."""
        frequency_grid = instrument_config.padded_frequency_grid_in_angstroms
        # Compute the CTF
        ctf_array = self.envelope(frequency_grid) * self.ctf(
            frequency_grid,
            wavelength_in_angstroms=instrument_config.wavelength_in_angstroms,
            defocus_offset=defocus_offset,
        )
        # ... compute the contrast as the CTF multiplied by the exit plane
        # phase shifts
        fourier_contrast_at_detector_plane = ctf_array * fourier_phase_at_exit_plane

        return fourier_contrast_at_detector_plane



"""
The image configuration and utility manager.
"""

import math
from functools import cached_property
from typing import Any, Callable, Optional, Union

import jax.numpy as jnp
from equinox import Module
from jaxtyping import Array, Float

from .._errors import error_if_not_positive
from ..constants import convert_keV_to_angstroms
from ..coordinates import make_coordinate_grid, make_frequency_grid
from ..image import (
    crop_to_shape,
    pad_to_shape,
    resize_with_crop_or_pad,
)


class InstrumentConfig(Module, strict=True):
    """Configuration and utilities for an electron microscopy image."""

    shape: tuple[int, int]
    pixel_size: Float[Array, ""]
    voltage_in_kilovolts: Float[Array, ""]
    electrons_per_angstrom_squared: Float[Array, ""]

    padded_shape: tuple[int, int]
    pad_mode: Union[str, Callable]

    def __init__(
        self,
        shape: tuple[int, int],
        pixel_size: float | Float[Array, ""],
        voltage_in_kilovolts: float | Float[Array, ""],
        electrons_per_angstrom_squared: float | Float[Array, ""] = 100.0,
        padded_shape: Optional[tuple[int, int]] = None,
        *,
        pad_scale: float = 1.0,
        pad_mode: Union[str, Callable] = "constant",
    ):
        """**Arguments:**

        - `shape`:
            Shape of the imaging plane in pixels.
        - `pixel_size`:
            The pixel size of the image in angstroms.
        - `voltage_in_kilovolts`:
            The incident energy of the electron beam.
        - `electrons_per_angstrom_squared`:
            The integrated dose rate of the electron beam.
        - `padded_shape`:
            The shape of the image after padding. If this argument is
            not given, it can be set by the `pad_scale` argument.
        - `pad_scale`: A scale factor at which to pad the image. This is
                       optionally used to set `padded_shape` and must be
                       greater than `1`. If `padded_shape` is set, this
                       argument is ignored.
        - `pad_mode`:
            The method of image padding. By default, `"constant"`.
            For all options, see `jax.numpy.pad`.
        """
        self.shape = shape
        self.pixel_size = error_if_not_positive(jnp.asarray(pixel_size))
        self.voltage_in_kilovolts = error_if_not_positive(
            jnp.asarray(voltage_in_kilovolts)
        )
        self.electrons_per_angstrom_squared = error_if_not_positive(
            jnp.asarray(electrons_per_angstrom_squared)
        )
        self.pad_mode = pad_mode
        # Set shape after padding
        if padded_shape is None:
            self.padded_shape = (int(pad_scale * shape[0]), int(pad_scale * shape[1]))
        else:
            self.padded_shape = padded_shape

    def __check_init__(self):
        if self.padded_shape[0] < self.shape[0] or self.padded_shape[1] < self.shape[1]:
            raise AttributeError(
                "ImageConfig.padded_shape is less than ImageConfig.shape in one or "
                "more dimensions."
            )

    @property
    def wavelength_in_angstroms(self) -> Float[Array, ""]:
        """The incident electron wavelength corresponding to the beam
        energy `voltage_in_kilovolts`.
        """
        return convert_keV_to_angstroms(self.voltage_in_kilovolts)

    @property
    def wavenumber_in_inverse_angstroms(self) -> Float[Array, ""]:
        """The incident electron wavenumber corresponding to the beam
        energy `voltage_in_kilovolts`.
        """
        return 2 * jnp.pi / self.wavelength_in_angstroms

    @cached_property
    def coordinate_grid_in_pixels(
        self,
    ) -> Float[Array, "{self.y_dim} {self.x_dim} 2"]:
        """A spatial coordinate system for the `shape`."""
        return make_coordinate_grid(shape=self.shape)

    @cached_property
    def coordinate_grid_in_angstroms(
        self,
    ) -> Float[Array, "{self.y_dim} {self.x_dim} 2"]:
        """Convenience property for `pixel_size * coordinate_grid_in_pixels`"""
        return _safe_multiply_by_constant(self.coordinate_grid_in_pixels, self.pixel_size)

    @cached_property
    def frequency_grid_in_pixels(
        self,
    ) -> Float[Array, "{self.y_dim} {self.x_dim//2+1} 2"]:
        """A spatial frequency coordinate system for the `shape`,
        with hermitian symmetry.
        """
        return make_frequency_grid(shape=self.shape)

    @cached_property
    def frequency_grid_in_angstroms(
        self,
    ) -> Float[Array, "{self.y_dim} {self.x_dim//2+1} 2"]:
        """Convenience property for `frequency_grid_in_pixels / pixel_size`"""
        return _safe_multiply_by_constant(
            self.frequency_grid_in_pixels, 1 / self.pixel_size
        )

    @cached_property
    def full_frequency_grid_in_pixels(
        self,
    ) -> Float[Array, "{self.y_dim} {self.x_dim} 2"]:
        """A spatial frequency coordinate system for the `shape`,
        without hermitian symmetry.
        """
        return make_frequency_grid(shape=self.shape, half_space=False)

    @cached_property
    def full_frequency_grid_in_angstroms(
        self,
    ) -> Float[Array, "{self.y_dim} {self.x_dim} 2"]:
        """Convenience property for `full_frequency_grid_in_pixels / pixel_size`"""
        return _safe_multiply_by_constant(
            self.full_frequency_grid_in_pixels, 1 / self.pixel_size
        )

    @cached_property
    def padded_coordinate_grid_in_pixels(
        self,
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim} 2"]:
        """A spatial coordinate system for the `padded_shape`."""
        return make_coordinate_grid(shape=self.padded_shape)

    @cached_property
    def padded_coordinate_grid_in_angstroms(
        self,
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim} 2"]:
        """Convenience property for `pixel_size * padded_coordinate_grid_in_pixels`"""
        return _safe_multiply_by_constant(
            self.padded_coordinate_grid_in_pixels, self.pixel_size
        )

    @cached_property
    def padded_frequency_grid_in_pixels(
        self,
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim//2+1} 2"]:
        """A spatial frequency coordinate system for the `padded_shape`,
        with hermitian symmetry.
        """
        return make_frequency_grid(shape=self.padded_shape)

    @cached_property
    def padded_frequency_grid_in_angstroms(
        self,
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim//2+1} 2"]:
        """Convenience property for `padded_frequency_grid_in_pixels / pixel_size`"""
        return _safe_multiply_by_constant(
            self.padded_frequency_grid_in_pixels, 1 / self.pixel_size
        )

    @cached_property
    def padded_full_frequency_grid_in_pixels(
        self,
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim} 2"]:
        """A spatial frequency coordinate system for the `padded_shape`,
        without hermitian symmetry.
        """
        return make_frequency_grid(shape=self.padded_shape, half_space=False)

    @cached_property
    def padded_full_frequency_grid_in_angstroms(
        self,
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim} 2"]:
        """Convenience property for `padded_full_frequency_grid_in_pixels / pixel_size`"""
        return _safe_multiply_by_constant(
            self.padded_full_frequency_grid_in_pixels, 1 / self.pixel_size
        )

    def crop_to_shape(
        self, image: Float[Array, "y_dim x_dim"]
    ) -> Float[Array, "{self.y_dim} {self.x_dim}"]:
        """Crop an image to `shape`."""
        return crop_to_shape(image, self.shape)

    def pad_to_padded_shape(
        self, image: Float[Array, "y_dim x_dim"], **kwargs: Any
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim}"]:
        """Pad an image to `padded_shape`."""
        return pad_to_shape(image, self.padded_shape, mode=self.pad_mode, **kwargs)

    def crop_or_pad_to_padded_shape(
        self, image: Float[Array, "y_dim x_dim"], **kwargs: Any
    ) -> Float[Array, "{self.padded_y_dim} {self.padded_x_dim}"]:
        """Reshape an image to `padded_shape` using cropping or padding."""
        return resize_with_crop_or_pad(
            image, self.padded_shape, mode=self.pad_mode, **kwargs
        )

    @property
    def n_pixels(self) -> int:
        """Convenience property for `math.prod(shape)`"""
        return math.prod(self.shape)

    @property
    def y_dim(self) -> int:
        """Convenience property for `shape[0]`"""
        return self.shape[0]

    @property
    def x_dim(self) -> int:
        """Convenience property for `shape[1]`"""
        return self.shape[1]

    @property
    def padded_y_dim(self) -> int:
        """Convenience property for `padded_shape[0]`"""
        return self.padded_shape[0]

    @property
    def padded_x_dim(self) -> int:
        """Convenience property for `padded_shape[1]`"""
        return self.padded_shape[1]

    @property
    def padded_n_pixels(self) -> int:
        """Convenience property for `math.prod(padded_shape)`"""
        return math.prod(self.padded_shape)


def _safe_multiply_by_constant(
    grid: Float[Array, "y_dim x_dim 2"], constant: Float[Array, ""]
) -> Float[Array, "y_dim x_dim 2"]:
    """Multiplies a coordinate grid by a constant in a
    safe way for gradient computation.
    """
    return jnp.where(grid != 0.0, jnp.asarray(constant) * grid, 0.0)



"""
Abstraction of a biological assembly. This assembles a structure
by computing a batch of subunits, parameterized by some geometry.
"""

from abc import abstractmethod
from functools import cached_property
from typing import Optional
from typing_extensions import override

import equinox as eqx
import jax
from equinox import AbstractVar
from jaxtyping import Array, Float

from ...rotations import SO3
from .._pose import AbstractPose
from .._structural_ensemble import (
    AbstractConformationalVariable,
    AbstractStructuralEnsemble,
    AbstractStructuralEnsembleBatcher,
)


class AbstractAssembly(AbstractStructuralEnsembleBatcher, strict=True):
    """Abstraction of a biological assembly.

    To subclass an `AbstractAssembly`,
        1) Overwrite the `AbstractAssembly.n_subunits`
           property
        2) Overwrite the `AbstractAssembly.positions`
           and `AbstractAssembly.rotations` properties.
    """

    subunit: AbstractVar[AbstractStructuralEnsemble]
    pose: AbstractVar[AbstractPose]
    conformation: AbstractVar[Optional[AbstractConformationalVariable]]

    n_subunits: AbstractVar[int]

    def __check_init__(self):
        if self.conformation is not None and self.subunit.conformation is None:
            # Make sure that if conformation is set, subunit is an AbstractEnsemble
            raise AttributeError(
                f"If {type(self)}.conformation is set, "
                "{type(self)}.subunit.conformation cannot be `None`."
            )
        if self.conformation is not None and self.subunit.conformation is not None:
            # ... if it is an AbstractEnsemble, the AbstractConformation must be the
            #  right type
            if not isinstance(self.conformation, type(self.subunit.conformation)):
                raise AttributeError(
                    f"{type(self)}.conformation must be type "
                    f" {type(self.subunit.conformation)} if {type(self)}.subunit is "
                    f"type {type(self.subunit)}."
                )

    @cached_property
    @abstractmethod
    def offsets_in_angstroms(self) -> Float[Array, "{n_subunits} 3"]:
        """The positions of each subunit."""
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def rotations(self) -> SO3:
        """The relative rotations between subunits."""
        raise NotImplementedError

    @cached_property
    def poses(self) -> AbstractPose:
        """
        Draw the poses of the subunits in the lab frame, measured
        from the rotation relative to the first subunit.
        """
        # Transform the subunit positions by the center of mass pose of the assembly.
        transformed_positions = (
            self.pose.rotate_coordinates(self.offsets_in_angstroms, inverse=False)
            + self.pose.offset_in_angstroms
        )
        # Transform the subunit rotations by the center of mass pose of the assembly.
        # This operation left multiplies by the pose rotation matrix, taking care that
        # first subunits are rotated to the center of mass frame, then the lab frame.
        transformed_rotations = jax.vmap(
            lambda com_rotation, subunit_rotation: com_rotation @ subunit_rotation,
            in_axes=[None, 0],
        )(self.pose.rotation, self.rotations)
        # Construct the batch of `AbstractPose`s
        cls = type(self.pose)
        make_assembly_poses = jax.vmap(
            lambda rot, pos: cls.from_rotation_and_translation(rot, pos)
        )

        return make_assembly_poses(transformed_rotations, transformed_positions)

    @cached_property
    def subunits(self) -> AbstractStructuralEnsemble:
        """Draw a realization of all of the subunits in the lab frame."""
        # Compute a list of subunits, configured at the correct conformations
        if self.subunit.conformation is not None:
            where = lambda s: (s.conformation, s.pose)
            return eqx.tree_at(where, self.subunit, (self.conformation, self.poses))
        else:
            where = lambda s: s.pose
            return eqx.tree_at(where, self.subunit, self.poses)

    @override
    def get_batched_structural_ensemble(self) -> AbstractStructuralEnsemble:
        return self.subunits



from . import (
    coordinates as coordinates,
    data as data,
    image as image,
    inference as inference,
    rotations as rotations,
    simulator as simulator,
)
from ._filter_specs import get_filter_spec as get_filter_spec
from ._filtered_transformations import (
    filter_grad_with_spec as filter_grad_with_spec,
    filter_value_and_grad_with_spec as filter_value_and_grad_with_spec,
    filter_vmap_with_spec as filter_vmap_with_spec,
)
from .cryojax_version import __version__ as __version__



from .atom_potential_integrator import (
    GaussianMixtureProjection as GaussianMixtureProjection,
)
from .base_potential_integrator import (
    AbstractPotentialIntegrator as AbstractPotentialIntegrator,
    AbstractVoxelPotentialIntegrator as AbstractVoxelPotentialIntegrator,
)
from .fourier_voxel_extract import (
    AbstractFourierVoxelExtraction as AbstractFourierVoxelExtraction,
    EwaldSphereExtraction as EwaldSphereExtraction,
    FourierSliceExtraction as FourierSliceExtraction,
)
from .nufft_project import (
    NufftProjection as NufftProjection,
)



"""
Methods for integrating the scattering potential directly onto the exit plane.
"""

from abc import abstractmethod
from typing import Generic, Optional, TypeVar

import jax.numpy as jnp
from equinox import AbstractVar, error_if, Module
from jaxtyping import Array, Complex

from ...image import maybe_rescale_pixel_size
from .._instrument_config import InstrumentConfig
from .._potential_representation import AbstractVoxelPotential


PotentialT = TypeVar("PotentialT")
VoxelPotentialT = TypeVar("VoxelPotentialT", bound="AbstractVoxelPotential")


class AbstractPotentialIntegrator(Module, Generic[PotentialT], strict=True):
    """Base class for a method of integrating a potential onto
    the exit plane.
    """

    @abstractmethod
    def compute_fourier_integrated_potential(
        self,
        potential: PotentialT,
        instrument_config: InstrumentConfig,
    ) -> Complex[
        Array, "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}"
    ]:
        raise NotImplementedError


class AbstractVoxelPotentialIntegrator(
    AbstractPotentialIntegrator[VoxelPotentialT], strict=True
):
    """Base class for a method of integrating a voxel-based potential."""

    pixel_rescaling_method: AbstractVar[Optional[str]]

    def _convert_raw_image_to_integrated_potential(
        self,
        fourier_integrated_potential_without_postprocess,
        potential,
        instrument_config,
    ):
        """Return the integrated potential in fourier space at the
        `instrument_config.pixel_size` and the `instrument_config.padded_shape.`
        """
        if self.pixel_rescaling_method is None:
            fourier_integrated_potential = error_if(
                potential.voxel_size * fourier_integrated_potential_without_postprocess,
                ~jnp.isclose(potential.voxel_size, instrument_config.pixel_size),
                f"Tried to use {type(self).__name__} with `{type(potential).__name__}."
                "voxel_size != InstrumentConfig.pixel_size`. If this is true, then "
                f"`{type(self).__name__}.pixel_rescaling_method` must not be set to "
                f"`None`. Try setting `{type(self).__name__}.pixel_rescaling_method = "
                "'bicubic'`.",
            )
            return fourier_integrated_potential
        else:
            fourier_integrated_potential = maybe_rescale_pixel_size(
                potential.voxel_size * fourier_integrated_potential_without_postprocess,
                potential.voxel_size,
                instrument_config.pixel_size,
                is_real=False,
                shape_in_real_space=instrument_config.padded_shape,
            )
            return fourier_integrated_potential



from abc import abstractmethod

import equinox as eqx

from .base_ensemble import AbstractStructuralEnsemble


class AbstractStructuralEnsembleBatcher(eqx.Module, strict=True):
    """A batching utility for structural ensembles."""

    @abstractmethod
    def get_batched_structural_ensemble(self) -> AbstractStructuralEnsemble:
        raise NotImplementedError



from .base_scattering_theory import AbstractScatteringTheory as AbstractScatteringTheory
from .common_functions import (
    compute_phase_shifts_from_integrated_potential as compute_phase_shifts_from_integrated_potential,  # noqa: E501
)
from .weak_phase_scattering_theory import (
    AbstractWeakPhaseScatteringTheory as AbstractWeakPhaseScatteringTheory,
    LinearSuperpositionScatteringTheory as LinearSuperpositionScatteringTheory,
    WeakPhaseScatteringTheory as WeakPhaseScatteringTheory,
)



from typing import Optional
from typing_extensions import override

import jax
import jax.numpy as jnp
from jaxtyping import Array, Complex, Float

from ...coordinates import make_1d_coordinate_grid
from ...image import downsample_to_shape_with_fourier_cropping, rfftn
from .._instrument_config import InstrumentConfig
from .._potential_representation import (
    GaussianMixtureAtomicPotential,
    PengAtomicPotential,
)
from .base_potential_integrator import AbstractPotentialIntegrator


class GaussianMixtureProjection(
    AbstractPotentialIntegrator[GaussianMixtureAtomicPotential | PengAtomicPotential],
    strict=True,
):
    upsampling_factor: Optional[int]

    def __init__(self, *, upsampling_factor: Optional[int] = None):
        """**Arguments:**

        - `upsampling_factor`:
            The factor by which to upsample the computation of the images.
            If `upsampling_factor` is greater than 1, the images will be computed
            at a higher resolution and then downsampled to the original resolution.
            This can be useful for reducing aliasing artifacts in the images.
        """  # noqa: E501
        self.upsampling_factor = upsampling_factor

    def __check_init__(self):
        if self.upsampling_factor is not None and self.upsampling_factor < 1:
            raise AttributeError(
                "`GaussianMixtureProjection.upsampling_factor` must "
                f"be greater than `1`. Got a value of {self.upsampling_factor}."
            )

    @override
    def compute_fourier_integrated_potential(
        self,
        potential: GaussianMixtureAtomicPotential | PengAtomicPotential,
        instrument_config: InstrumentConfig,
    ) -> Complex[
        Array, "{instrument_config.padded_y_dim} {instrument_config.padded_x_dim//2+1}"
    ]:
        """Compute a projection from the atomic potential and transform it to Fourier
        space.

        **Arguments:**

        - `potential`: The atomic potential to project.
        - `instrument_config`: The configuration of the imaging instrument.

        **Returns:**

        The Fourier transform of the integrated potential.
        """  # noqa: E501

        if self.upsampling_factor is not None:
            pixel_size = instrument_config.pixel_size / self.upsampling_factor
            shape = (
                instrument_config.padded_y_dim * self.upsampling_factor,
                instrument_config.padded_x_dim * self.upsampling_factor,
            )
        else:
            pixel_size = instrument_config.pixel_size
            shape = instrument_config.padded_shape

        grid_x = make_1d_coordinate_grid(shape[1], pixel_size)
        grid_y = make_1d_coordinate_grid(shape[0], pixel_size)

        if isinstance(potential, PengAtomicPotential):
            if potential.b_factors is None:
                gaussian_widths = potential.scattering_factor_b
            else:
                gaussian_widths = (
                    potential.scattering_factor_b + potential.b_factors[:, None]
                )

            gaussian_amplitudes = potential.scattering_factor_a

        elif isinstance(potential, GaussianMixtureAtomicPotential):
            gaussian_amplitudes = potential.gaussian_amplitudes
            gaussian_widths = potential.gaussian_widths

        else:
            raise ValueError(
                "Supported types for `potential` are `PengAtomicPotential` and "
                " `GaussianMixtureAtomicPotential`."
            )

        projection = _evaluate_2d_real_space_gaussian(
            grid_x, grid_y, potential.atom_positions, gaussian_amplitudes, gaussian_widths
        )

        if self.upsampling_factor is not None:
            fourier_projection = downsample_to_shape_with_fourier_cropping(
                projection,
                downsampled_shape=instrument_config.padded_shape,
                get_real=False,
            )
        else:
            fourier_projection = rfftn(projection)

        return fourier_projection


@jax.jit
def _evaluate_2d_real_space_gaussian(
    grid_x: Float[Array, " x_dim"],
    grid_y: Float[Array, " y_dim"],
    atom_positions: Float[Array, "n_atoms 3"],
    a: Float[Array, "n_atoms n_gaussians_per_atom"],
    b: Float[Array, "n_atoms n_gaussians_per_atom"],
) -> Float[Array, "y_dim x_dim"]:
    """Evaluate a gaussian on a 3D grid.

    **Arguments:**

    - `grid_x`: The x-coordinates of the grid.
    - `grid_y`: The y-coordinates of the grid.
    - `pos`: The center of the gaussian.
    - `a`: A scale factor.
    - `b`: The scale of the gaussian.

    **Returns:**

    The potential of the gaussian on the grid.
    """

    b_inverse = 4.0 * jnp.pi / b

    gauss_x = (
        jnp.exp(
            -jnp.pi
            * b_inverse[None, :, :]
            * ((grid_x[:, None] - atom_positions.T[0, :]) ** 2)[:, :, None]
        )
        * a[None, :, :]
        * b_inverse[None, :, :]
    )
    gauss_y = jnp.exp(
        -jnp.pi
        * b_inverse[None, :, :]
        * ((grid_y[:, None] - atom_positions.T[1, :]) ** 2)[:, :, None]
    )

    gauss_x = jnp.transpose(gauss_x, (2, 1, 0))
    gauss_y = jnp.transpose(gauss_y, (2, 0, 1))

    image = 4 * jnp.pi * jnp.sum(jnp.matmul(gauss_y, gauss_x), axis=0)

    return image



"""Routines for downsampling arrays"""

from typing import overload

import jax.numpy as jnp
from jaxtyping import Array, Inexact

from ._edges import crop_to_shape
from ._fft import fftn, ifftn


@overload
def downsample_with_fourier_cropping(
    image_or_volume: Inexact[Array, "_ _ _"],
    downsampling_factor: float | int,
    get_real: bool = True,
) -> Inexact[Array, "_ _ _"]: ...


@overload
def downsample_with_fourier_cropping(
    image_or_volume: Inexact[Array, "_ _"],
    downsampling_factor: float | int,
    get_real: bool = True,
) -> Inexact[Array, "_ _"]: ...


def downsample_with_fourier_cropping(
    image_or_volume: Inexact[Array, "_ _"] | Inexact[Array, "_ _ _"],
    downsampling_factor: float | int,
    get_real: bool = True,
) -> Inexact[Array, "_ _"] | Inexact[Array, "_ _ _"]:
    """Downsample an array using fourier cropping.

    **Arguments:**

    - `image_or_volume`: The image or volume array to downsample.
    - `downsample_factor`:
        A scale factor at which to downsample `image_or_volume`
        by. Must be a value greater than `1`.
    - `get_real`:
        If `False`, the `image_or_volume` is returned in fourier space.

    **Returns:**

    The downsampled `image_or_volume`, with shape reduced by a factor of
    `downsample_factor`.
    """
    downsampling_factor = float(downsampling_factor)
    if downsampling_factor < 1.0:
        raise ValueError("`downsampling_factor` must be greater than 1.0")
    if image_or_volume.ndim == 2:
        image = image_or_volume
        new_shape = (
            int(image.shape[0] / downsampling_factor),
            int(image.shape[1] / downsampling_factor),
        )
        downsampled_array = downsample_to_shape_with_fourier_cropping(
            image, new_shape, get_real=get_real
        )
    elif image_or_volume.ndim == 3:
        volume = image_or_volume
        new_shape = (
            int(volume.shape[0] / downsampling_factor),
            int(volume.shape[1] / downsampling_factor),
            int(volume.shape[2] / downsampling_factor),
        )
        downsampled_array = downsample_to_shape_with_fourier_cropping(
            volume, new_shape, get_real=get_real
        )
    else:
        raise ValueError(
            "`downsample_with_fourier_cropping` can only crop images and volumes. "
            f"Got an array with number of dimensions {image_or_volume.ndim}."
        )

    if get_real:
        return (
            downsampled_array.real
            if jnp.issubdtype(image_or_volume.dtype, jnp.floating)
            else downsampled_array
        )
    else:
        return downsampled_array


@overload
def downsample_to_shape_with_fourier_cropping(
    image_or_volume: Inexact[Array, "_ _"],
    downsampled_shape: tuple[int, int],
    get_real: bool = True,
) -> Inexact[Array, "_ _"]: ...


@overload
def downsample_to_shape_with_fourier_cropping(
    image_or_volume: Inexact[Array, "_ _ _"],
    downsampled_shape: tuple[int, int, int],
    get_real: bool = True,
) -> Inexact[Array, "_ _ _"]: ...


def downsample_to_shape_with_fourier_cropping(
    image_or_volume: Inexact[Array, "_ _"] | Inexact[Array, "_ _ _"],
    downsampled_shape: tuple[int, int] | tuple[int, int, int],
    get_real: bool = True,
) -> Inexact[Array, "_ _"] | Inexact[Array, "_ _ _"]:
    """Downsample an array to a specified shape using fourier cropping.

    **Arguments:**

    - `image_or_volume`: The image or volume array to downsample.
    - `downsampled_shape`:
        The new shape after fourier cropping.
    - `get_real`:
        If `False`, the `image_or_volume` is returned in fourier space.

    **Returns:**

    The downsampled `image_or_volume`, at the new real-space shape
    `downsampled_shape`. If `get_real = False`, return
    the downsampled array in fourier space assuming hermitian symmetry,
    with the zero frequency component in the corner.
    """
    fourier_array = jnp.fft.fftshift(fftn(image_or_volume))
    cropped_fourier_array = crop_to_shape(fourier_array, downsampled_shape)
    if get_real:
        return ifftn(jnp.fft.ifftshift(cropped_fourier_array))
    else:
        return jnp.fft.ifftshift(cropped_fourier_array)[
            ..., : downsampled_shape[-1] // 2 + 1
        ]



"""
Abstractions of ensembles of biological specimen.
"""

from abc import abstractmethod
from typing import Any, Optional
from typing_extensions import override

from equinox import AbstractVar, Module

from .._pose import AbstractPose, EulerAnglePose
from .._potential_representation import AbstractPotentialRepresentation
from .base_conformation import AbstractConformationalVariable


class AbstractStructuralEnsemble(Module, strict=True):
    """A map from a pose and conformational variable to an `AbstractPotential`."""

    conformational_space: AbstractVar[Any]
    pose: AbstractVar[AbstractPose]
    conformation: AbstractVar[Optional[AbstractConformationalVariable]]

    @abstractmethod
    def get_potential_at_conformation(self) -> AbstractPotentialRepresentation:
        """Get the scattering potential in the center of mass
        frame."""
        raise NotImplementedError

    def get_potential_in_lab_frame(self) -> AbstractPotentialRepresentation:
        """Get the scattering potential in the lab frame."""
        potential = self.get_potential_at_conformation()
        return potential.rotate_to_pose(self.pose)


class SingleStructureEnsemble(AbstractStructuralEnsemble, strict=True):
    """An "ensemble" with one conformation."""

    conformational_space: AbstractPotentialRepresentation
    pose: AbstractPose
    conformation: None

    def __init__(
        self,
        conformational_space: AbstractPotentialRepresentation,
        pose: Optional[AbstractPose] = None,
    ):
        """**Arguments:**

        - `conformational_space`: The scattering potential representation of the
                         specimen as a single scattering potential object.
        - `pose`: The pose of the specimen.
        """
        self.conformational_space = conformational_space
        self.pose = pose or EulerAnglePose()
        self.conformation = None

    @override
    def get_potential_at_conformation(self) -> AbstractPotentialRepresentation:
        """Get the scattering potential in the center of mass
        frame.
        """
        return self.conformational_space



"""
Read and write atomic models in various formats.
Large amounts of the code are adapted from the ioSPI package
"""

import pathlib
from typing import Literal, overload

import gemmi
import numpy as np
from jaxtyping import Float, Int

from ._gemmi import (
    center_gemmi_model,
    clean_gemmi_structure,
    extract_atom_b_factors,
    extract_atom_positions_and_numbers,
    extract_gemmi_atoms,
)


@overload
def read_atoms_from_pdb(
    filename: str | pathlib.Path,
    i_model: int = 0,
    clean: bool = True,
    center: bool = True,
    assemble: bool = True,
    get_b_factors: Literal[False] = False,
) -> tuple[Float[np.ndarray, "n_atoms 3"], Int[np.ndarray, " n_atoms"]]: ...


@overload
def read_atoms_from_pdb(
    filename: str | pathlib.Path,
    i_model: int = 0,
    clean: bool = True,
    center: bool = True,
    assemble: bool = True,
    get_b_factors: Literal[True] = True,
) -> tuple[
    Float[np.ndarray, "n_atoms 3"],
    Int[np.ndarray, " n_atoms"],
    Float[np.ndarray, " n_atoms"],
]: ...


def read_atoms_from_pdb(
    filename: str | pathlib.Path,
    i_model: int = 0,
    clean: bool = True,
    center: bool = True,
    assemble: bool = True,
    get_b_factors: bool = False,
) -> (
    tuple[Float[np.ndarray, "n_atoms 3"], Int[np.ndarray, " n_atoms"]]
    | tuple[
        Float[np.ndarray, "n_atoms 3"],
        Int[np.ndarray, " n_atoms"],
        Float[np.ndarray, " n_atoms"],
    ]
):
    """Read atomic information from a PDB file using `gemmi`.

    **Arguments:**

    - `path`: Path to PDB file.
    - `i_model`: Index of the returned model in the `gemmi` Structure.
    - `clean`: If `True`, use `gemmi` remove_* methods to clean up structure.
    - `center`: If `True`, center the model so that its center of mass coincides
             with the origin.
    - `assemble`: If `True`, use `gemmi` make_assembly to build biological object.

    **Returns:**

    A tuple whose first element is a `numpy` array of coordinates containing
    atomic positions, and whose second element is an array of atomic element
    numbers. To be clear,

    ```python
    atom_positons, atom_element_numbers = read_atoms_from_pdb(...)
    ```

    Notes
    -----
    Currently Hydrogen atoms are not read in!
    We should look into adding hydrogens: does this slow things down
    appreciably?  Also, does it have a big effect on the scattering?
    """
    if pathlib.Path(filename).suffix != ".pdb":
        raise IOError(
            "Tried to read PDB file, but the filename does not have a .pdb "
            f"suffix. Got filename '{filename}'."
        )
    structure = gemmi.read_structure(str(filename))
    return _read_atom_info_from_structure(
        structure,
        i_model=i_model,
        clean=clean,
        center=center,
        assemble=assemble,
        get_b_factors=get_b_factors,
    )


@overload
def read_atoms_from_cif(
    filename: str | pathlib.Path,
    i_model: int = 0,
    clean: bool = True,
    center: bool = True,
    assemble: bool = True,
    get_b_factors: Literal[False] = False,
) -> tuple[Float[np.ndarray, "n_atoms 3"], Int[np.ndarray, " n_atoms"]]: ...


@overload
def read_atoms_from_cif(
    filename: str | pathlib.Path,
    i_model: int = 0,
    clean: bool = True,
    center: bool = True,
    assemble: bool = True,
    get_b_factors: Literal[True] = True,
) -> tuple[
    Float[np.ndarray, "n_atoms 3"],
    Int[np.ndarray, " n_atoms"],
    Float[np.ndarray, " n_atoms"],
]: ...


def read_atoms_from_cif(
    filename: str | pathlib.Path,
    i_model: int = 0,
    clean: bool = True,
    center: bool = True,
    assemble: bool = True,
    get_b_factors: bool = False,
) -> (
    tuple[Float[np.ndarray, "n_atoms 3"], Int[np.ndarray, " n_atoms"]]
    | tuple[
        Float[np.ndarray, "n_atoms 3"],
        Int[np.ndarray, " n_atoms"],
        Float[np.ndarray, " n_atoms"],
    ]
):
    """Read atomic information from an mmCIF file using `gemmi`.

    **Arguments:**

    - `path`: Path to mmCIF file.
    - `i_model`: Index of the returned model in the `gemmi` Structure.
    - `clean`: If `True`, use `gemmi` remove_* methods to clean up structure.
    - `center`: If `True`, center the model so that its center of mass coincides
             with the origin.
    - `assemble`: If `True`, use `gemmi` make_assembly to build biological object.

    **Returns:**

    A tuple whose first element is a `numpy` array of coordinates containing
    atomic positions, and whose second element is an array of atomic element
    numbers. To be clear,

    ```python
    atom_positons, atom_element_numbers = read_atoms_from_pdb(...)
    ```
    """
    if pathlib.Path(filename).suffix != ".cif":
        raise IOError(
            "Tried to read mmCIF file, but the filename does not have a .cif "
            f"suffix. Got filename '{filename}'."
        )
    cif_block = gemmi.cif.read(str(filename))[0]
    structure = gemmi.make_structure_from_block(cif_block)
    return _read_atom_info_from_structure(
        structure,
        i_model=i_model,
        clean=clean,
        center=center,
        assemble=assemble,
        get_b_factors=get_b_factors,
    )


def _read_atom_info_from_structure(
    structure,
    i_model: int = 0,
    clean: bool = True,
    center: bool = True,
    assemble: bool = True,
    get_b_factors: bool = False,
):
    atoms = _read_atoms_from_structure(
        structure, i_model=i_model, clean=clean, center=center, assemble=assemble
    )
    atom_positions, atom_element_numbers = extract_atom_positions_and_numbers(atoms)

    if not get_b_factors:
        return atom_positions, atom_element_numbers
    else:
        atom_b_factors = extract_atom_b_factors(atoms)
        return atom_positions, atom_element_numbers, atom_b_factors


def _read_atoms_from_structure(
    structure,
    i_model: int = 0,
    clean: bool = True,
    center: bool = True,
    assemble: bool = True,
):
    if clean:
        structure = clean_gemmi_structure(structure)
    if center:
        model = center_gemmi_model(structure[i_model])
    model = structure[i_model]
    if assemble:
        assembly = structure.assemblies[i_model]
        chain_naming = gemmi.HowToNameCopiedChain.AddNumber
        model = gemmi.make_assembly(assembly, model, chain_naming)

    atoms = extract_gemmi_atoms(model)

    return atoms



import jax.random as jr
import numpy as np
import pytest

from cryojax.image import downsample_with_fourier_cropping


@pytest.mark.parametrize("downsampling_factor", [1.0, 2.0, 4.0])
def test_rescale_by_downsampling_preserves_sum(downsampling_factor):
    rng_key = jr.PRNGKey(seed=1234)
    random = 2.0 + 1.0 * jr.normal(rng_key, (100, 100))
    downsampled_random = downsample_with_fourier_cropping(random, downsampling_factor)
    np.testing.assert_allclose(random.sum(), downsampled_random.sum())
